from matplotlib import pyplot as plt
import scipy.stats
from scipy.sparse import spmatrix, issparse, csr_matrix
from scipy.stats import multivariate_normal
from anndata import AnnData
from typing import Optional, Union
from shapely.geometry import Point, MultiPoint
from numpy.linalg import svd, solve, lstsq
from numpy.random import randn
from cytocraft.craft import *
from cytocraft.model import BasisShapeModel


def DeriveRotation(W, X, Mask):
    F = int(W.shape[0] / 2)
    Rotation = np.zeros((F, 3, 3))
    # I = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -1]])
    for i in range(F):
        # filter genes
        Wi = W[:, Mask[i, :]]
        # filter cells
        Wi_filter = Wi[~np.isnan(Wi).any(axis=1), :]
        while Wi_filter.shape[0] < 6:
            Mask[i, :] = change_last_true(Mask[i, :])
            Wi = W[:, Mask[i, :]]
            # filter cells
            Wi_filter = Wi[~np.isnan(Wi).any(axis=1), :]
        idx = int(find_subarray(Wi_filter, Wi[i * 2]) / 2)
        Xi = X[Mask[i, :], :]
        model = factor(Wi_filter)
        R = kabsch_numpy(np.dot(model.Rs[idx], model.Ss[0]).T, Xi)[0]
        Rotation[i] = R
    return Rotation


# def UpdateF(RM, W):
#    F = int(W.shape[0] / 2)
#    for j in range(W.shape[1]):
#        a1 = b1 = c1 = d1 = a2 = b2 = c2 = d2 = a3 = b3 = c3 = d3 = 0
#        for i in range(F):
#            if not (
#                np.isnan(W[i * 2 : i * 2 + 2, j]).any() or np.isnan(RM[i, :, :]).any()
#            ):
#                a1 += RM[i, 0, 0] * RM[i, 0, 0] + RM[i, 0, 1] * RM[i, 0, 1]
#                b1 += RM[i, 0, 0] * RM[i, 1, 0] + RM[i, 0, 1] * RM[i, 1, 1]
#                c1 += RM[i, 0, 0] * RM[i, 2, 0] + RM[i, 0, 1] * RM[i, 2, 1]
#
#                a2 += RM[i, 1, 0] * RM[i, 0, 0] + RM[i, 1, 1] * RM[i, 0, 1]
#                b2 += RM[i, 1, 0] * RM[i, 1, 0] + RM[i, 1, 1] * RM[i, 1, 1]
#                c2 += RM[i, 1, 0] * RM[i, 2, 0] + RM[i, 1, 1] * RM[i, 2, 1]
#
#                a3 += RM[i, 2, 0] * RM[i, 0, 0] + RM[i, 2, 1] * RM[i, 0, 1]
#                b3 += RM[i, 2, 0] * RM[i, 1, 0] + RM[i, 2, 1] * RM[i, 1, 1]
#                c3 += RM[i, 2, 0] * RM[i, 2, 0] + RM[i, 2, 1] * RM[i, 2, 1]
#
#                d1 += RM[i, 0, 0] * W[i * 2, j] + RM[i, 0, 1] * W[i * 2 + 1, j]
#                d2 += RM[i, 1, 0] * W[i * 2, j] + RM[i, 1, 1] * W[i * 2 + 1, j]
#                d3 += RM[i, 2, 0] * W[i * 2, j] + RM[i, 2, 1] * W[i * 2 + 1, j]
#            else:
#                # print("skip cell" + str(i) + "for gene" + str(j))
#                pass
#
#        args = np.array([[a1, b1, c1], [a2, b2, c2], [a3, b3, c3]])
#        results = np.array([d1, d2, d3])
#        newXi = LA.solve(args, results)
#        try:
#            newX = np.append(
#                newX,
#                [newXi],
#                axis=0,
#            )
#        except NameError:
#            newX = np.array([newXi])
#
#    return newX


def write_sim_pdb(simX, prefix="simchain", outpath="./Results/"):
    show_shape = simX.T
    with open(outpath + "/" + prefix + ".pdb", "w") as out:
        out.write(
            "HEADER".ljust(10, " ")
            + "CHROMOSOMES".ljust(40, " ")
            + "21-FEB-23"
            + " "
            + "CMB1".ljust(4, " ")
            + "\n"
        )
        out.write(
            "TITLE".ljust(10, " ") + "CHROMOSOMES OF MICE BRAIN INFERED FROM sST DATA\n"
        )
        i = 1
        j = 1
        chain = 1
        Is = []
        atom_lines = []
        for idx in range(show_shape.shape[1]):
            atom_lines.append(
                "ATOM".ljust(6, " ")
                + str(i).rjust(5, " ")
                + " "
                + "O".ljust(3, " ")
                + " "
                + "GLY".ljust(3, " ")
                + " "
                + chr(64 + chain)
                + str(j).rjust(4, " ")
                + " "
                + str("%.3f" % (show_shape[0, idx])).rjust(8, " ")
                + str("%.3f" % (show_shape[1, idx])).rjust(8, " ")
                + str("%.3f" % (show_shape[2, idx])).rjust(8, " ")
                + "2.00".rjust(6, " ")
                + "10.00".rjust(6, " ")
                + " "
                + "O".rjust(2, " ")
                + "\n"
            )
            Is.append(i)
            i += 1
            j += 1
        atom_lines.append(
            "TER".ljust(6, " ")
            + str(i).rjust(5, " ")
            + " "
            + "".ljust(3, " ")
            + " "
            + "GLY".ljust(3, " ")
            + " "
            + chr(64 + chain)
            + str(j - 1).rjust(4, " ")
            + "\n"
        )
        chain = 1
        connect_lines = []
        for l in Is:
            if l - 1 in Is and l + 1 in Is:
                connect_lines.append(
                    "CONECT".ljust(6, " ")
                    + str(l).rjust(5, " ")
                    + str(l - 1).rjust(5, " ")
                    + str(l + 1).rjust(5, " ")
                    + "\n"
                )
            elif not l - 1 in Is:
                connect_lines.append(
                    "CONECT".ljust(6, " ")
                    + str(l).rjust(5, " ")
                    + str(l + 1).rjust(5, " ")
                    + "\n"
                )
            elif not l + 1 in Is:
                connect_lines.append(
                    "CONECT".ljust(6, " ")
                    + str(l).rjust(5, " ")
                    + str(l - 1).rjust(5, " ")
                    + "\n"
                )
        out.writelines(atom_lines)
        out.writelines(connect_lines)
        out.write("END\n")


def write_row_to_csv(filename, row):
    # open the file in append mode
    with open(filename, "a", newline="") as f:
        # create the csv writer
        writer = csv.writer(f)
        # write a row to the csv file
        writer.writerow(row)


def scale_X(X, size=1):
    # calculate the minimum and maximum values along each axis
    min_x, min_y, min_z = X.min(axis=0)
    max_x, max_y, max_z = X.max(axis=0)
    # calculate the range of values along each axis
    range_x = max_x - min_x
    range_y = max_y - min_y
    range_z = max_z - min_z
    # calculate the scale factor to fit the coordinates within the size
    scale = size / max(range_x, range_y, range_z)
    # normalize the coordinates by subtracting the minimum values and multiplying by the scale factor
    scaled_X = (X - [min_x, min_y, min_z]) * scale
    # return the scaled coordinates
    return scaled_X, scale


def euclidean_distance_3d_matrix(X):
    import math

    # coords is a list of tuples of the form (x, y, z)
    # returns a numpy array of shape (len(coords), len(coords)) where the element at (i, j) is the distance between coords[i] and coords[j]
    n = X.shape[0]
    matrix = np.zeros((n, n))

    for i in range(n):
        for j in range(i + 1, n):
            x1, y1, z1 = X[i, :]
            x2, y2, z2 = X[j, :]
            d = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2 + (z2 - z1) ** 2)
            matrix[i][j] = d
            matrix[j][i] = d
    return matrix


def distance_to_similarity(matrix):
    # matrix is a numpy array of shape (n, n) where the element at (i, j) is the distance between two points
    # returns a numpy array of shape (n, n) where the element at (i, j) is the similarity between two points
    # similarity is defined as 1 / (1 + distance)
    similarity = 1 / (1 + matrix)
    return similarity


def centerX(X):
    """
    X = (nGene,3)
    """
    if len(X.shape) == 2 and X.shape[1] == 3:
        mean_x, mean_y, mean_z = X.mean(axis=0)
        centered_X = X - [mean_x, mean_y, mean_z]
        return centered_X, [mean_x, mean_y, mean_z]
    else:
        print("X shape error!")


def main():
    # Read auguments
    GenomeStructure = pd.read_csv(sys.argv[1], delimiter="\t")
    sample_name = os.path.basename(sys.argv[1]).split(".")[0]
    Ngene = int(sys.argv[2])  ### simulated gene number
    Ncell = int(sys.argv[3])  ### simulated cell number
    rateCap = float(sys.argv[4])  ### set capture rate
    rateDrop = float(sys.argv[5])  ### set gene loss rate
    resolution = int(sys.argv[6])  ### difine the resolution
    Ngene_for_rotation_derivation = int(sys.argv[7])
    noise = int(sys.argv[8])
    mode = sys.argv[9].strip()
    outpath = sys.argv[10]
    csv = sys.argv[11]

    # Define outpath
    outpath = (
        outpath
        + "/"
        + sample_name
        + "_NG"
        + str(Ngene)
        + "_NC"
        + str(Ncell)
        + "_RC"
        + str(rateCap)
        + "_RD"
        + str(rateDrop)
        + "_RS"
        + str(resolution)
        + "_NGF"
        + str(Ngene_for_rotation_derivation)
        + "_NS"
        + str(noise)
        + "_"
        + mode
    )
    # Exit if the experiment has been done
    # result = pd.read_csv(csv, sep=",")
    # repeats = result.query(
    #     f"GeneNumber == {Ngene} and \\"
    #     f"CellNumber == {Ncell} and \\"
    #     f"CaptureRate == {rateCap} and \\"
    #     f"DropRate == {rateDrop} and \\"
    #     f"Resolution== {resolution} and \\"
    #     f"NumberofGeneforFactorization == {Ngene_for_rotation_derivation} and \\"
    #     f"Noise == {noise} and \\"
    #     f"Mode == '%s'" % mode
    # )
    # if len(repeats) >= 10:
    #     sys.exit(1)
    Path(outpath).mkdir(parents=True, exist_ok=True)
    # Check the number of arguments and set random seed
    if len(sys.argv) == 13:
        # Try to convert the 12th argument to an integer
        try:
            seed = int(sys.argv[12])
        except ValueError:
            # If the argument is not a valid integer, print an error message and exit
            print("Invalid seed argument. Please provide an integer.")
            sys.exit(1)
    elif len(sys.argv) == 12:
        # If the argument is not given, generate a random seed
        seed = random.randint(0, 1000000)
    else:
        print(
            "Argument length is not correct, please check each parameter for correctness."
        )
        sys.exit(1)
    random.seed(seed)
    np.random.seed(seed)

    if mode == "continous":
        Conformation = np.array(GenomeStructure[["x", "y", "z"]].head(Ngene))
    elif mode == "random":
        sampled_indices = sorted(
            np.random.choice(len(GenomeStructure), size=Ngene, replace=False)
        )
        Conformation = np.array(GenomeStructure[["x", "y", "z"]].iloc[sampled_indices])
    else:
        print("input mode is not correct")
    TID = generate_id()

    # start logging
    # stdout = sys.stdout
    # log_file = open(outpath + "/" + TID + ".log", "w")
    # sys.stdout = log_file
    print(f"The seed is {seed}.")

    ##### GENERATE SIMULATION DATA
    ScaledCoords, _ = scale_X(Conformation, resolution)
    simX, _ = centerX(ScaledCoords)
    # save_data(simX, outpath + "/"+TID+"_simX_resolution" + str(resolution) + ".pkl")

    #### generate random RM
    randRM = generate_random_rotation_matrices(Ncell)
    save_data(
        randRM,
        outpath
        + "/"
        + TID
        + "_randRM_nCell"
        + str(Ncell)
        + "_rateCap"
        + str(rateCap)
        + "_rateDrop"
        + str(rateDrop)
        + "_resolution"
        + str(resolution)
        + ".pkl",
    )

    #### generate W
    simW = np.zeros((Ncell * 2, Ngene))
    for c in range(randRM.shape[0]):
        XY = np.dot(simX, randRM[c, :, :])[:, 0:2]
        # XY = np.dot(randRM[c, :, :], simX.T).T[:, 0:2]
        simW[c * 2] = XY[:, 0]
        simW[c * 2 + 1] = XY[:, 1]

    # save_data(simW, outpath + "/HMEC_simW_" + str(Ncell) + ".pkl")

    # add Noise1: random noise on locs
    simW = np.random.normal(simW, noise)
    print("noise: " + str(noise))

    #### expression matrix
    UBx = int(np.max(simW[::2], axis=1).max())
    LBx = int(np.min(simW[::2], axis=1).min())
    UBy = int(np.max(simW[1::2], axis=1).max())
    LBy = int(np.min(simW[1::2], axis=1).min())

    Matrix = np.empty((Ncell, Ngene, UBx - LBx, UBy - LBy), dtype="i1")
    x = np.arange(LBx, UBx)  # create a 1D array of x values
    y = np.arange(LBy, UBy)  # create a 1D array of y values
    xx, yy = np.meshgrid(x, y)  # create a 2D grid of x and y values
    xy = np.stack((xx, yy), axis=-1)  # create a 3D array of (x,y) pairs

    # Creat fixed Gaussian noise 3 matrix
    # noise_scale = int(noise / 2)
    # noise_matrix = np.random.normal(0, noise_scale, (UBx - LBx, UBy - LBy))
    if noise == 0:
        sigma = np.eye(2) * 8
    else:
        # add Noise2: random sigma
        A = np.random.rand(2, 2)  # generate a 2x2 random matrix
        sigma = (
            np.dot(A, A.transpose()) * 8
        )  # create a symmetric, positive semi-definite matrix for the covariance
    for c in range(Ncell):
        for g in range(Ngene):
            loc = simW[c * 2 : (c + 1) * 2, g]
            # add random noise 2 on exp matrix
            exp = (
                multivariate_normal.pdf(xy, mean=loc, cov=sigma, allow_singular=True)
                * 128
            )  # compute the multivariate normal pdf on the grid
            # add fixed noise 3 to exp matrix
            Matrix[c, g] = (
                exp.T.round().clip(min=0, max=127).astype("i1")
            )  # round and cast to integer

    #### drop counts based on capture rate
    CapMatrix = np.random.binomial(
        Matrix, rateCap
    )  # generate random numbers from a binomial distribution with Matrix as the number of trials and rateCap as the probability

    #### drop centers based on drop rate
    mask = (
        np.random.random((Ncell, Ngene)) < rateDrop
    )  # generate a boolean mask with rateDrop as the probability
    CapMatrix[mask] = (
        0  # set the elements of CapMatrix that correspond to True in the mask to zero
    )

    #### write gem as data frame
    columns = ["geneID", "x", "y", "MIDCount", "ExonCount", "CellID"]
    rows = []

    for index, n in np.ndenumerate(CapMatrix):
        if n > 0:
            row = [
                "gene" + str(index[1]),
                str(index[2]),
                str(index[3]),
                str(n),
                str(n),
                str(index[0]),
            ]
            rows.append(row)

    gem = pd.DataFrame(rows, columns=columns)

    ##### RUN Cytocraft
    sample_name = "HMEC"

    # read input gem
    models = []
    X = pd.Series(
        sorted(gem.geneID.drop_duplicates(), key=lambda fname: int(fname.strip("gene")))
    )
    Ngene = len(X)
    print(
        "TASK ID: "
        + TID
        + "\n"
        + "Cell Number: "
        + str(Ncell)
        + "\n"
        + "Gene Number: "
        + str(Ngene)
        + "\n"
        + "Gene Number for Rotation Derivatives: "
        + str(Ngene_for_rotation_derivation)
        + "\n"
        + "rateCap: "
        + str(rateCap)
        + "\n"
        + "rateDrop: "
        + str(rateDrop)
        + "\n"
        + "resolution: "
        + str(resolution)
        + "\n"
        + "mode: "
        + mode
        + "\n"
    )

    ##### update X
    W = get_centers(gem, gem.CellID.drop_duplicates().values, X)
    W = normalizeW(W)
    # get rotation R through shared X and input W
    RM = generate_random_rotation_matrices(int(W.shape[0] / 2))
    CellUIDs = list(gem.CellID.drop_duplicates())
    Mask = MASK(gem, GeneIDs=X, CellIDs=CellUIDs, Ngene=Ngene_for_rotation_derivation)
    X, _, _ = UpdateF(RM, W, X)
    write_sim_pdb(
        scale_X(X, 0.5)[0],
        prefix=TID
        + "_top100_nCell"
        + str(Ncell)
        + "_rateCap"
        + str(rateCap)
        + "_rateDrop"
        + str(rateDrop)
        + "_resolution"
        + str(resolution)
        + "_initial",
        outpath=outpath,
    )
    try:
        for loop in range(30):
            RM = DeriveRotation(W, X, Mask)
            try:
                X, _, _ = UpdateF(RM, W, X)
            except np.linalg.LinAlgError:
                return "numpy.linalg.LinAlgError"

            # test
            # save_data(W, outpath + "/HMEC_W_" + "loop_" + str(loop) + ".pkl")
            # save_data(X, outpath + "/HMEC_X_" + "loop_" + str(loop) + ".pkl")
            # save_data(RM, outpath + "/HMEC_RM_" + "loop_" + str(loop) + ".pkl")

            # evaluate
            rmsd1 = kabsch_numpy(
                normalizeF(simX, method="mean"), normalizeF(X, method="mean")
            )[2]
            X_mirror = np.copy(X)
            X_mirror[:, 2] = -X_mirror[:, 2]
            rmsd2 = kabsch_numpy(
                normalizeF(simX, method="mean"), normalizeF(X_mirror, method="mean")
            )[2]
            minrmsd = min(rmsd1, rmsd2)
            print(
                "Distance between ground truth and reconstructed structure for loop "
                + str(loop + 1)
                + " is: "
                + str(rmsd1)
                + " and "
                + str(rmsd2)
            )
            write_sim_pdb(
                scale_X(X, 0.5)[0],
                prefix=TID
                + "_top100_nCell"
                + str(Ncell)
                + "_rateCap"
                + str(rateCap)
                + "_rateDrop"
                + str(rateDrop)
                + "_resolution"
                + str(resolution)
                + "_updated"
                + str(loop + 1)
                + "times",
                outpath=outpath,
            )
            if minrmsd < 0.005:
                break
    except Exception as error:
        print(error)
        row = (
            TID,
            sample_name,
            str(Ngene),
            str(Ncell),
            str(rateCap),
            str(rateDrop),
            str(resolution),
            str(Ngene_for_rotation_derivation),
            str(noise),
            mode,
            "NA",
            "NA",
            "NA",
        )
        if not csv == None:
            write_row_to_csv(csv, row)
        sys.exit(1)

    ####### evaluation
    if minrmsd == rmsd1:
        mirror = 0
    else:
        mirror = 1
    D = euclidean_distance_3d_matrix(X)
    S = distance_to_similarity(D)
    D_ = euclidean_distance_3d_matrix(simX)
    S_ = distance_to_similarity(D_)
    print(
        "RMSD Distance".ljust(18, " ")
        + "\t"
        + "Mirror"
        + "\t"
        + "Spearman Correlation Coefficient"
    )
    print(
        str(minrmsd)
        + "\t"
        + str(mirror)
        + "\t"
        + str(scipy.stats.spearmanr(S_.flatten(), S.flatten())[0])
    )

    # stop logging
    # sys.stdout = stdout
    # log_file.close()

    row = (
        TID,
        sample_name,
        str(Ngene),
        str(Ncell),
        str(rateCap),
        str(rateDrop),
        str(resolution),
        str(Ngene_for_rotation_derivation),
        str(noise),
        mode,
        str(minrmsd),
        str(mirror),
        str(scipy.stats.spearmanr(S_.flatten(), S.flatten())[0]),
    )

    if not csv == None:
        write_row_to_csv(csv, row)


if __name__ == "__main__":
    main()



import sys, os, argparse, pickle, copy, random, string, csv, warnings
import numpy as np
import pandas as pd
import scanpy as sc
from tqdm import tqdm
from pathlib import Path
from scipy import linalg as LA
from scipy.linalg import LinAlgWarning
from scipy.spatial.transform import Rotation as R
import scipy.cluster.hierarchy as sch
from cytocraft import model
from cytocraft import util
from cytocraft.stereopy import *
from cytocraft.rigid import *
import warnings
import traceback

warnings.filterwarnings("ignore")

"""
This module implements main functions.
"""


def get_centers(gem, CellIDs, GeneIDs):
    """
    Calculate the transcription centers of genes for a given set of cells.

    Args:
        gem (DataFrame): The gene expression matrix.
        CellIDs (list): The list of cell IDs to consider.
        GeneIDs (list): The list of gene IDs to calculate the centers for.

    Returns:
        numpy.ndarray: The observed matrix of transcription centers.

    """
    Z = np.zeros((len(CellIDs) * 2, len(GeneIDs)))
    i = 0
    for c in CellIDs:
        j = 0
        # subset the gem in terms of individual cells
        gem_cell = gem[gem.CellID == c]
        for n in GeneIDs:
            if n not in gem_cell.geneID.values:
                x_median = np.nan
                y_median = np.nan
            else:
                gem_cell_gene = gem_cell[gem_cell.geneID == n]
                x_median = np.average(
                    gem_cell_gene.x.values.astype(int),
                    weights=gem_cell_gene.MIDCount.values.astype(float),
                )
                y_median = np.average(
                    gem_cell_gene.y.values.astype(int),
                    weights=gem_cell_gene.MIDCount.values.astype(float),
                )
            Z[i * 2, j] = x_median
            Z[i * 2 + 1, j] = y_median
            j += 1
        i += 1
    return Z


def MASK(gem, GeneIDs, CellIDs, Ngene):
    mask = np.zeros((len(CellIDs), len(GeneIDs)))
    i = 0
    for c in CellIDs:
        gem_cell = gem[gem.CellID == c]
        GeneUIDs_cell = (
            gem_cell.groupby(["geneID"])["MIDCount"]
            .count()
            .reset_index(name="Count")
            .sort_values(["Count"], ascending=False)
            .geneID[:Ngene]
        )

        mask[i] = np.isin(GeneIDs, GeneUIDs_cell)
        i += 1
    return np.bool_(mask)


def DeriveRotation(Z, F, Mask, CellUIDs, adata):
    """
    R-step: Derives rotation matrices for each cell based on the given inputs.

    Args:
        Z (numpy.ndarray): The input array of shape (N, M) representing gene expression levels.
        F (numpy.ndarray): The input array of shape (N, 3) representing cell positions.
        Mask (numpy.ndarray): The input array of shape (Ncell, M) representing the gene mask for each cell.
        CellUIDs (list): The list of unique identifiers for each cell.
        adata (pandas.DataFrame): The input dataframe containing additional cell data.

    Returns:
        Rotation (numpy.ndarray): The array of shape (Ncell, 3, 3) representing the rotation matrices for each cell.
        Z (numpy.ndarray): The updated array of shape (N', M) after removing cells.
        CellUIDs (list): The updated list of unique identifiers after removing cells.
        adata (pandas.DataFrame): The updated dataframe after removing cells.
    """
    Ncell = int(Z.shape[0] / 2)
    Rotation = np.zeros((Ncell, 3, 3))
    pop_indices = []
    for i in range(Ncell):
        try:
            # print("derive rotation for cell " + str(i))
            # select genes
            Zi = Z[:, Mask[i, :]]
            # filter cells
            Zi_filter = Zi[~np.isnan(Zi).any(axis=1), :]
            while Zi_filter.shape[0] < 6:
                # reduce the numebr of gene in Mask if cell number less than 3
                Mask[i, :] = change_last_true(Mask[i, :])
                # print("reduce one gene for cell " + str(i))
                # filter genes
                Zi = Z[:, Mask[i, :]]
                # filter cells
                Zi_filter = Zi[~np.isnan(Zi).any(axis=1), :]
            idx = int(find_subarray(Zi_filter, Zi[i * 2]) / 2)
            Fi = F[Mask[i, :], :]
            model = factor(Zi_filter)
            R = kabsch_numpy(np.dot(model.Rs[idx], model.Ss[0]).T, Fi)[0]
            Rotation[i] = R
        except Exception as err:
            print("Cell ID [" + str(CellUIDs[i]) + "] is removed due to: ", err)
            pop_indices.append(i)
    for i in sorted(pop_indices, reverse=True):
        Rotation = np.delete(Rotation, obj=i, axis=0)
        Z = np.delete(Z, obj=i * 2 + 1, axis=0)
        Z = np.delete(Z, obj=i * 2, axis=0)
        adata = adata[~(adata.obs.index.values == str(CellUIDs[i])), :]
        del CellUIDs[i]
    return Rotation, Z, CellUIDs, adata


def UpdateF(RM, Z, GeneUID, *F_old):
    """
    F-step: Update the coordinates of configuration based on the given rotation matrices RM and observation matrix of transcription centers Z.

    Parameters:
    - RM (numpy.ndarray): A 3-dimensional matrix representing the rotation matrices.
    - Z (numpy.ndarray): A 2-dimensional matrix representing the observation matrix of transcription centers.
    - GeneUID (list): A list of gene unique identifiers for transcription centers.
    - F_old (numpy.ndarray): The old coordinates of configuration.

    Returns:
    - newF (numpy.ndarray): A 2-dimensional matrix representing the updated coordinates of configuration.
    - Z (numpy.ndarray): A 2-dimensional matrix representing the filtered observation matrix of transcription centers Z after removing some genes.
    - GeneUID (list): A list of gene unique identifiers after removing some genes.
    - F_old (numpy.ndarray): A 2-dimensional matrix representing the filtered old coordinates of configuration after removing some genes.

    Note:
    - If F_old is not provided, the function will only update newF, Z, and GeneUID.
    - If F_old is provided, the function will update F_old as well.
    """
    Ncell = int(Z.shape[0] / 2)
    pop_indices = []
    for j in range(Z.shape[1]):
        a1 = b1 = c1 = d1 = a2 = b2 = c2 = d2 = a3 = b3 = c3 = d3 = 0
        for i in range(Ncell):
            if not (
                np.isnan(Z[i * 2 : i * 2 + 2, j]).any() or np.isnan(RM[i, :, :]).any()
            ):
                a1 += RM[i, 0, 0] * RM[i, 0, 0] + RM[i, 0, 1] * RM[i, 0, 1]
                b1 += RM[i, 0, 0] * RM[i, 1, 0] + RM[i, 0, 1] * RM[i, 1, 1]
                c1 += RM[i, 0, 0] * RM[i, 2, 0] + RM[i, 0, 1] * RM[i, 2, 1]

                a2 += RM[i, 1, 0] * RM[i, 0, 0] + RM[i, 1, 1] * RM[i, 0, 1]
                b2 += RM[i, 1, 0] * RM[i, 1, 0] + RM[i, 1, 1] * RM[i, 1, 1]
                c2 += RM[i, 1, 0] * RM[i, 2, 0] + RM[i, 1, 1] * RM[i, 2, 1]

                a3 += RM[i, 2, 0] * RM[i, 0, 0] + RM[i, 2, 1] * RM[i, 0, 1]
                b3 += RM[i, 2, 0] * RM[i, 1, 0] + RM[i, 2, 1] * RM[i, 1, 1]
                c3 += RM[i, 2, 0] * RM[i, 2, 0] + RM[i, 2, 1] * RM[i, 2, 1]

                d1 += RM[i, 0, 0] * Z[i * 2, j] + RM[i, 0, 1] * Z[i * 2 + 1, j]
                d2 += RM[i, 1, 0] * Z[i * 2, j] + RM[i, 1, 1] * Z[i * 2 + 1, j]
                d3 += RM[i, 2, 0] * Z[i * 2, j] + RM[i, 2, 1] * Z[i * 2 + 1, j]
            else:
                # print("skip cell" + str(i) + "for gene" + str(j))
                pass

        args = np.array([[a1, b1, c1], [a2, b2, c2], [a3, b3, c3]])
        results = np.array([d1, d2, d3])
        try:
            newFi = LA.solve(args, results)
            try:
                newF = np.append(
                    newF,
                    [newFi],
                    axis=0,
                )
            except NameError:
                newF = np.array([newFi])
        except (np.linalg.LinAlgError, LinAlgWarning) as err:
            print("Gene No.[" + str(j + 1) + "] is removed due to: ", err)
            pop_indices.append(j)

    # Drop genes
    ## F_old also need to drop genes, so that F can be compared
    if F_old:
        F_old = F_old[0]
        for i in sorted(pop_indices, reverse=True):
            del GeneUID[i]
            Z = np.delete(Z, obj=i, axis=1)
            F_old = np.delete(F_old, obj=i, axis=0)
        return newF, Z, GeneUID, F_old
    else:
        for i in sorted(pop_indices, reverse=True):
            del GeneUID[i]
            Z = np.delete(Z, obj=i, axis=1)
        return newF, Z, GeneUID


 def kabsch_numpy(in_crds1, in_crds2, center=False):
    """
    Returns rmsd and optional rotation between 2 sets of [nx3] arrays.

    This requires numpy for svd decomposition.
    The transform direction: transform(m, ref_crd) => target_crd.
    """

    crds1 = np.array(in_crds1)
    crds2 = np.array(in_crds2)
    assert crds1.shape[1] == 3
    assert crds1.shape == crds2.shape
    
    # Compute centroids
    centroid_P = np.mean(crds1, axis=0)
    centroid_Q = np.mean(crds2, axis=0)

    # Optimal translation
    t = centroid_Q - centroid_P

    # Center the points
    if center:
        crds1 = crds1 - centroid_P
        crds2 = crds2 - centroid_Q

    n_vec = np.shape(crds1)[0]
    # compute the covariance matrix
    correlation_matrix = np.dot(np.transpose(crds1), crds2)
    # SVD decomposition
    v, s, w = np.linalg.svd(correlation_matrix)
    is_reflection = (np.linalg.det(v) * np.linalg.det(w)) < 0.0
    if is_reflection:
        s[-1] = -s[-1]

    E0 = sum(sum(crds1 * crds1)) + sum(sum(crds2 * crds2))
    rmsd_sq = (E0 - 2.0 * sum(s)) / float(n_vec)
    rmsd_sq = max([rmsd_sq, 0.0])
    rmsd = np.sqrt(rmsd_sq)

    if is_reflection:
        v[-1, :] = -v[-1, :]

    # optimal rotation
    rot33 = np.dot(v, w).transpose()
    # print(is_reflection)
    return rot33, t, rmsd, is_reflection


#def kabsch_numpy(P, Q, center=False):
#    """
#    Computes the optimal rotation and translation to align two sets of points (P -> Q),
#    and their RMSD.
#
#    :param P: A Nx3 matrix of points
#    :param Q: A Nx3 matrix of points
#    :param center: Whether to center the points
#    :return: A tuple containing the optimal rotation matrix, the optimal
#             translation vector, and the RMSD.
#    """
#    P = np.array(P)
#    Q = np.array(Q)
#    assert P.shape == Q.shape, "Matrix dimensions must match"
#
#    # Compute centroids
#    centroid_P = np.mean(P, axis=0)
#    centroid_Q = np.mean(Q, axis=0)
#
#    # Optimal translation
#    t = centroid_Q - centroid_P
#
#    # Center the points
#    if center:
#        P = P - centroid_P
#        Q = Q - centroid_Q
#
#    # Compute the covariance matrix
#    H = np.dot(P.T, Q)
#
#    # SVD
#    U, S, Vt = np.linalg.svd(H)
#
#    # Validate right-handed coordinate system
#    is_reflection = np.linalg.det(np.dot(Vt.T, U.T)) < 0.0
#    if is_reflection:
#        Vt[-1, :] *= -1.0
#
#    # Optimal rotation
#    R = np.dot(Vt.T, U.T)
#
#    # RMSD
#    rmsd = np.sqrt(np.sum(np.square(np.dot(P, R.T) - Q)) / P.shape[0])
#
#    return R, t, rmsd, is_reflection


def normalizeZ(Z):
    normZ = np.empty_like(Z)
    for i in range(int(Z.shape[0])):
        normZ[i] = Z[i] - np.nanmean(Z[i])
    return normZ


def filterGenes(array, Genes, threshold):
    # threshold is a float between 0 and 1, indicating the maximum proportion of np.nans allowed in a column
    # returns a new array with only the columns that have less than threshold proportion of np.nans
    # if array is empty or threshold is invalid, returns None

    # check if array is empty
    if array.size == 0:
        return None

    # check if threshold is valid
    if not (0 <= threshold <= 1):
        return None

    # get the number of rows and columns in the array
    rows, cols = array.shape

    # create a list to store the indices of the columns to keep
    keep_cols = []

    # loop through each column
    for i in range(cols):
        # get the column as a 1D array
        col = array[:, i]

        # count the number of np.nans in the column
        nan_count = np.count_nonzero(np.isnan(col))

        # calculate the proportion of np.nans in the column
        nan_prop = nan_count / rows

        # if the proportion is less than the threshold, add the index to the list
        if nan_prop < threshold:
            keep_cols.append(i)

    # create a new array with only the columns in the list
    new_array = array[:, keep_cols]
    Genes = list(np.array(Genes)[keep_cols])

    # return the new array
    return new_array, Genes


def get_gene_chr(species):
    if species == "Mouse" or species == "Mice":
        gtf_file = (
            os.path.dirname(os.path.realpath(__file__))
            + "/gtf/gencode.vM34.annotation.gene.gtf"
        )
    elif species == "Axolotls" or species == "Axolotl":
        gtf_file = (
            os.path.dirname(os.path.realpath(__file__))
            + "/gtf/AmexT_v47-AmexG_v6.0-DD.gene.gtf"
        )
    elif species == "Human" or species == "Monkey":
        gtf_file = (
            os.path.dirname(os.path.realpath(__file__))
            + "/gtf/gencode.v44.basic.annotation.gene.gtf"
        )
    gene_chr = {}
    with open(gtf_file, "r") as f:
        for line in f:
            if not line.startswith("#"):
                inf = line.strip().split("\t")
                if inf[2] == "gene":
                    chrom = inf[0]
                    if chrom not in gene_chr.keys():
                        gene_chr[chrom] = []
                    gene_names = inf[8].split(";")
                    try:
                        if (
                            species == "Mouse"
                            or species == "Mice"
                            or species == "Human"
                            or species == "Monkey"
                        ):
                            gene_symbol = gene_names[2].split()[1].strip('"')
                        elif species == "Axolotls" or species == "Axolotl":
                            gene_symbol = gene_names[1].split()[1].strip('"')
                        gene_chr[gene_symbol] = chrom
                    except IndexError as e:
                        print(
                            f"{type(e).__name__} at line {e.__traceback__.tb_lineno} of {__file__}: {e}"
                        )
                        pass
    return gene_chr


def gene_gene_distance_matrix(F):
    GeneList = list(F.index)
    N = len(GeneList)
    DM = np.zeros((N, N))
    for n, _ in enumerate(GeneList):
        for m, _ in enumerate(GeneList[: n + 1]):
            d = np.linalg.norm(F.iloc[n] - F.iloc[m])
            DM[n, m] = DM[m, n] = d
    return DM


def RMSD_distance_matrix(
    Confs, GeneLists, order=None, ngene=100, compare_method="pair", norm_method=None
):
    # check if Confs and GeneLists have the same keys
    if order == None:
        keys = list(Confs.keys())
    else:
        keys = order
    for i in keys:
        if not i in GeneLists.keys():
            print(f"Warning: {i} in Confs but not in GeneLists")
            return

    # calculate the distance matrix
    N = len(keys)
    DM = np.zeros((N, N))
    if compare_method == "complete":
        from functools import reduce

        intersected_values = reduce(np.intersect1d, GeneLists.values())
        intersected_values = intersected_values[:ngene]
        if len(intersected_values) < ngene:
            print(
                f"Warning: {len(intersected_values)} common genes between samples are less than {ngene}"
            )
    from tqdm import tqdm

    for n, key_n in enumerate(tqdm(keys)):
        for m, key_m in enumerate(keys[: n + 1]):
            if compare_method == "pair":
                intersected_values = np.intersect1d(GeneLists[key_n], GeneLists[key_m])
                intersected_values = intersected_values[:ngene]
                if len(intersected_values) < ngene:
                    print(
                        f"Warning: {len(intersected_values)} common genes between {key_n} and {key_m} are less than {ngene}"
                    )
            boolean_arrays_n = np.in1d(GeneLists[key_n], intersected_values)
            boolean_arrays_m = np.in1d(GeneLists[key_m], intersected_values)
            Conf_n = Confs[key_n][boolean_arrays_n]
            Conf_m = Confs[key_m][boolean_arrays_m]
            if norm_method:
                Conf_n = normalizeF(Conf_n, method=norm_method)
                Conf_m = normalizeF(Conf_m, method=norm_method)
            rmsd1 = kabsch_numpy(Conf_n, Conf_m)[2]
            rmsd2 = kabsch_numpy(mirror(Conf_n), Conf_m)[2]
            DM[n, m] = DM[m, n] = min(rmsd1, rmsd2)
    return DM


def expression_similarity(adatas, ngene, method="pearson", compare_method="pair"):
    if method == "pearson":
        similarity_matrix = np.zeros((len(adatas), len(adatas)))
        for i, sample1 in enumerate(adatas.keys()):
            for j, sample2 in enumerate(adatas.keys()):
                if compare_method == "pair":
                    intersected_values = np.intersect1d(
                        GeneLists[key_n], GeneLists[key_m]
                    )
                    # print("number of common gene: " + str(len(intersected_values)))
                    intersected_values = intersected_values[:ngene]
                elif compare_method == "complete":
                    intersected_values = reduce(np.intersect1d, GeneLists.values())
                if len(intersected_values) < ngene:
                    print(
                        f"Warning: {len(intersected_values)} common genes between {sample1} and {sample2} are less than {ngene}"
                    )
                boolean_arrays_n = np.in1d(GeneLists[sample1], intersected_values)
                boolean_arrays_m = np.in1d(GeneLists[sample2], intersected_values)
                average_expression1 = np.mean(
                    adatas[sample1].X[:, boolean_arrays_n], axis=0
                )
                average_expression2 = np.mean(
                    adatas[sample2].X[:, boolean_arrays_m], axis=0
                )
                similarity = np.corrcoef(average_expression1, average_expression2)[0, 1]
                similarity_matrix[i, j] = similarity
        return similarity_matrix
    # elif method == "dendrogram":
    #    Exp = np.zeros((len(adatas), ngene))
    #    for i, sample in enumerate(adatas.keys()):
    #        if not compare_method == "complate":
    #            print(
    #                "only 'complete' compare method is supported if using 'dendrogram' similarity"
    #            )
    #        intersected_values = reduce(np.intersect1d, GeneLists.values())[:ngene]
    #        if len(intersected_values) < ngene:
    #            print(
    #                f"Warning: {len(intersected_values)} common genes between samples are less than {ngene}"
    #            )
    #        boolean_arrays = np.in1d(GeneLists[sample], intersected_values)
    #        Exp[i] = np.mean(adatas[sample].X[:, boolean_arrays], axis=0)
    #    similarity_matrix = dendrogram_similarity(
    #        Exp, method="average", metric="euclidean"
    #    )
    #    return similarity_matrix
    else:
        print("method not supported")
        return


def mirror(F):
    mirrorF = np.copy(F)
    mirrorF[:, 2] = -mirrorF[:, 2]
    return mirrorF


def euclidean_distance(coord1, coord2):
    distance = math.sqrt(sum([(x1 - x2) ** 2 for x1, x2 in zip(coord1, coord2)]))
    return distance


def knn_neighbors(X, k):
    if k != 0:
        knn_m = np.zeros((X.shape[0], X.shape[0]))
        for i in range(X.shape[0]):
            ids = np.argpartition(X[i], -k)[-k:]
            top_set = set(X[i, ids])
            if len(top_set) == 1:
                b = X[i] == top_set.pop()
                ids = []
                offset = 1
                left = True
                while (len(ids) < k) and (offset < X.shape[0]):
                    # while len(ids) < k:
                    if left:
                        idx = i + offset
                    else:
                        idx = i - offset
                    if idx < 0 or idx > len(b) - 1:
                        offset += 1
                        left = not left
                        continue
                    if b[idx]:
                        ids.append(idx)
                    offset += 1
                    left = not left
            knn_m[i, ids] = 1

        knn_m = (knn_m + knn_m.T) / 2
        knn_m[np.nonzero(knn_m)] = 1
    else:
        knn_m = X
    return knn_m


def write_pdb(show_F, genechr, geneLst, write_path, sp, seed, prefix="chain"):
    if sp == "Axolotls":
        uniquechains = [
            chr for chr in sorted(set(genechr.keys())) if chr.startswith("chr")
        ]
    else:
        uniquechains = sorted(set(genechr.keys()))

    # i is global gene index in genome include terminals
    # j is global gene index in genome
    i = 1
    j = 1
    chain_idx = 0
    for chain in uniquechains:
        k = 0
        show_shape = show_F.T
        rows = []
        rows.append(
            "HEADER".ljust(10, " ")
            + "CHROMOSOMES".ljust(40, " ")
            + get_date_today()
            + "   "
            + str(seed).ljust(4, " ")
            + "\n"
        )
        rows.append(
            "TITLE".ljust(10, " ") + "CHROMOSOME CONFORMATION INFERED FROM sST DATA\n"
        )
        Is = []
        for g in genechr[chain]:
            if g in geneLst:
                idx = geneLst.index(g)
                # normal process
                rows.append(
                    "ATOM".ljust(6, " ")
                    + str(i).rjust(5, " ")
                    + "  "
                    + "O".ljust(3, " ")
                    + " "
                    + "GLY".ljust(3, " ")
                    + " "
                    + chr(65 + chain_idx)
                    + str(j).rjust(4, " ")
                    + "    "
                    + str("%.3f" % (show_shape[0, idx])).rjust(8, " ")
                    + str("%.3f" % (show_shape[1, idx])).rjust(8, " ")
                    + str("%.3f" % (show_shape[2, idx])).rjust(8, " ")
                    + "2.00".rjust(6, " ")
                    + "10.00".rjust(6, " ")
                    + "          "
                    + "O".rjust(2, " ")
                    + "\n"
                )
                Is.append(i)
                i += 1
                j += 1
                k += 1
        if k == 0:
            continue
        i += 1
        rows.append(
            "TER".ljust(6, " ")
            + str(i - 1).rjust(5, " ")
            + "  "
            + "".ljust(3, " ")
            + " "
            + "GLY".ljust(3, " ")
            + " "
            + chr(65 + chain_idx)
            + str(j - 1).rjust(4, " ")
            + "\n"
        )
        for l in Is:
            if l - 1 in Is and l + 1 in Is:
                rows.append(
                    "CONECT".ljust(6, " ")
                    + str(l).rjust(5, " ")
                    + str(l - 1).rjust(5, " ")
                    + str(l + 1).rjust(5, " ")
                    + "\n"
                )
            elif not l - 1 in Is:
                rows.append(
                    "CONECT".ljust(6, " ")
                    + str(l).rjust(5, " ")
                    + str(l + 1).rjust(5, " ")
                    + "\n"
                )
            elif not l + 1 in Is:
                rows.append(
                    "CONECT".ljust(6, " ")
                    + str(l).rjust(5, " ")
                    + str(l - 1).rjust(5, " ")
                    + "\n"
                )
        rows.append("END\n")
        out = open(write_path + "/" + prefix + chain + ".pdb", "wb", 100 * (2**20))
        data = "".join(rows).encode()
        out.write(data)
        out.close()
        chain_idx += 1


def read_gem_as_csv(path, sep="\t"):
    gem = pd.read_csv(
        path,
        sep=sep,
        comment="#",
        dtype=str,
        converters={
            "x": float,
            "y": float,
            "cell_id": str,
            "cell": str,
            "CellID": str,
            "MIDCount": float,
            "MIDCounts": float,
            "expr": float,
        },
    )
    gem["x"] = gem["x"].astype(int)
    gem["y"] = gem["y"].astype(int)
    # cell column
    if "cell" in gem.columns:
        gem.rename(columns={"cell": "CellID"}, inplace=True)
    elif "cell_id" in gem.columns:
        gem.rename(columns={"cell_id": "CellID"}, inplace=True)
    # count column
    if "MIDCounts" in gem.columns:
        gem.rename(columns={"MIDCounts": "MIDCount"}, inplace=True)
    elif "expr" in gem.columns:
        gem.rename(columns={"expr": "MIDCount"}, inplace=True)
    # gene column
    if "gene" in gem.columns:
        gem.rename(columns={"gene": "geneID"}, inplace=True)
    elif "GeneID" in gem.columns:
        gem.rename(columns={"GeneID": "geneID"}, inplace=True)

    return gem


def read_gem_as_adata(gem_path, sep="\t", SN="adata"):
    data = read_gem(file_path=gem_path, bin_type="cell_bins", sep=sep)
    data.tl.raw_checkpoint()
    adata = stereo_to_anndata(
        data, flavor="scanpy", sample_id=SN, reindex=False, output=None
    )
    return adata


def split_gem(gem_path, celltype, ctkey, cikey, gsep):
    split_gem_paths = {}
    gem = read_gem_as_csv(gem_path, sep=gsep)
    for ct in celltype[ctkey].dropna().drop_duplicates():
        ct_legal = legalname(ct)
        gem_ct_path = gem_path + "." + ctkey + "." + ct_legal + ".tsv"
        split_gem_paths[ct] = gem_ct_path
        print("split gem path of " + ct + ": " + gem_ct_path)
        if cikey is not None:
            cellids = celltype[celltype[ctkey] == ct][cikey].values.astype(str)
        else:
            cellids = celltype[celltype[ctkey] == ct].index.values.astype(str)
        gem[gem.CellID.isin(cellids)].to_csv(gem_ct_path, sep="\t", index=False)
    return split_gem_paths


def read_gem_header(gem_path, sep="\t"):
    with open(gem_path, "r") as f:
        header = f.readline().strip().split(sep)
        return header


def check_gem_header(header):
    if (
        ("geneID" in header or "gene" in header or "GeneID" in header)
        and ("cell" in header or "CellID" in header or "cell_id" in header)
        and "x" in header
        and "y" in header
        and ("MIDCount" in header or "MIDCounts" in header or "expr" in header)
    ):
        return True
    else:
        return False


def change_last_true(arr):
    # Find the index of the last True value
    last_true_index = np.where(arr == True)[0][-1]

    # Set the value at that index to False
    arr[last_true_index] = False

    return arr


def find_subarray(arr1, arr2):
    n = arr1.shape[0]
    for i in range(n):
        if np.array_equal(arr1[i], arr2):
            return i
    return print("Error! Try to reduce Ngene in MASK function")


def generate_random_rotation_matrices(n):
    # use the Rotation.random method to generate n random rotations
    rotations = R.random(n)
    # convert the rotations to matrices
    matrices = rotations.as_matrix()
    # return the matrices
    return matrices


def get_date_today():
    from datetime import datetime

    # get the current date as a datetime object
    today = datetime.today()
    # format the date as DD-MM-YY
    date_string = today.strftime("%d-%b-%y")
    # print the date string
    return date_string


def load_data(file):
    with open(file, "rb") as f:
        x = pickle.load(f)
    return x


def save_data(data, file):
    with open(file, "wb") as f:
        pickle.dump(data, f)


def generate_id():
    return "".join(random.choices(string.ascii_uppercase + string.digits, k=4))


def legalname(string):
    import re

    # Replace spaces with underscores
    string = re.sub(r"\s+", "_", string)

    # Remove non-alphanumeric characters
    string = re.sub(r"[^a-zA-Z0-9_\-\.]", "", string)

    # Remove leading/trailing underscores
    string = re.sub(r"^_+|_+$", "", string)

    # Remove leading/trailing hyphens
    string = re.sub(r"^-+|-+$", "", string)

    # Remove leading/trailing periods
    string = re.sub(r"^\.+|\.+$", "", string)

    return string


def mean_radius(x):
    center = np.mean(x, axis=0)
    dist = np.linalg.norm(x - center, axis=1)
    return np.mean(dist)  # Return the average distance, i.e. the average radius


def normalizeF(x, method="mean"):
    if method == "L2norm":
        return x / (np.linalg.norm(x) ** 2)
    elif method == "mean":
        r = mean_radius(x)
        return x / r


def get_GeneUID(gem):
    return list(
        gem.groupby(["geneID"])["MIDCount"]
        .count()
        .reset_index(name="Count")
        .sort_values(["Count"], ascending=False)
        .geneID
    )


def get_CellUID(gem):
    return list(gem.CellID.drop_duplicates())


def run_craft(
    gem_path,
    species,
    seed,
    out_path,
    sep,
    threshold_for_gene_filter=0.9,
    threshold_for_rmsd=0.25,
    Ngene_for_rotation_derivation=None,
    percent_of_gene_for_rotation_derivation=None,
):
    # set seed
    random.seed(seed)
    np.random.seed(seed)
    #################### INIT ####################
    models = []
    SN = os.path.basename(os.path.splitext(gem_path)[0])
    TID = generate_id()
    outpath = out_path + "/" + SN + "_" + TID + "/"

    # make dir
    Path(outpath).mkdir(parents=True, exist_ok=True)

    # start logging
    log_file = open(outpath + "/" + SN + "_" + TID + ".log", "w")
    sys.stdout = log_file

    # read input gem
    gem = read_gem_as_csv(gem_path, sep=sep)
    adata = read_gem_as_adata(gem_path, sep=sep, SN=SN)
    GeneUIDs = get_GeneUID(gem)
    if Ngene_for_rotation_derivation is None:
        Ngene_for_rotation_derivation = int(
            float(percent_of_gene_for_rotation_derivation) * len(GeneUIDs)
        )
    try:
        adata = craft(
            gem=gem,
            adata=adata,
            species=species,
            nderive=Ngene_for_rotation_derivation,
            thresh=threshold_for_gene_filter,
            thresh_rmsd=threshold_for_rmsd,
            seed=seed,
            samplename=SN,
            outpath=outpath,
        )
    except Exception as e:
        print(
            f"{type(e).__name__} at line {e.__traceback__.tb_lineno} of {__file__}: {e}",
            file=sys.stderr,
        )
        print("conformation reconstruction failed.", file=sys.stderr)
        sys.stdout = sys.__stdout__
        log_file.close()  # stop logging
    else:
        print("conformation reconstruction finished.")
        adata.write_h5ad(filename=outpath + "adata.h5ad")
        sys.stdout = sys.__stdout__
        log_file.close()  # stop logging


def craft(
    gem,
    adata,
    species,
    nderive=10,
    thresh=0.9,
    thresh_rmsd=0.01,
    seed=999,
    samplename=None,
    outpath=False,
):
    """
    Perform the cytocraft algorithm to generate a 3D reconstruction of transcription centers.

    Parameters:
    - gem (numpy.ndarray): The gene expression matrix.
    - adata (AnnData): The annotated data object containing cell and gene information.
    - species (str): The species of the data.
    - nderive (int, optional): The number of genes used for rotation derivation. Default is 10.
    - thresh (float, optional): The threshold for gene filtering. Default is 0.9.
    - thresh_rmsd (float, optional): The threshold for convergence of the conformation. Default is 0.25.
    - seed (int, optional): The random seed for reproducibility. Default is 999.
    - samplename (str, optional): The name of the sample. Default is None.
    - outpath (bool or str, optional): The output path for writing PDB files. Default is False.

    Returns:
    - adata (AnnData): The annotated data object with the 3D reconstruction information.

    """
    # set seed
    random.seed(seed)
    np.random.seed(seed)
    TID = generate_id()
    # run cytocraft
    GeneUIDs = get_GeneUID(gem)
    CellUIDs = get_CellUID(gem)

    print("Speceis: " + species)
    if samplename:
        print("Sample Name: " + samplename)

    print(
        "Seed: "
        + str(seed)
        + "\n"
        + "Cell Number: "
        + str(len(CellUIDs))
        + "\n"
        + "Gene Number: "
        + str(len(GeneUIDs))
        + "\n"
        + "Cutoff for gene filter is: "
        + str(thresh)
        + "\n"
        + "Anchor Gene Number is: "
        + str(nderive)
        + "\n"
        + "Task ID: "
        + TID
        + "\n"
    )

    # reading corresponding gene annotation file as dictionary
    gene_chr = get_gene_chr(species)

    # generate and normalize observation TC matrix 'Z'
    Z = get_centers(gem, CellUIDs, GeneUIDs)
    Z, GeneUIDs = filterGenes(Z, GeneUIDs, threshold=thresh)
    Z = normalizeZ(Z)
    # generate random Rotation Matrix 'R'
    RM = generate_random_rotation_matrices(int(Z.shape[0] / 2))
    F, Z, GeneUIDs = UpdateF(RM, Z, GeneUIDs)
    if outpath is not False:
        write_pdb(
            F,
            gene_chr,
            GeneUIDs,
            write_path=outpath,
            sp=species,
            seed=seed,
            prefix=samplename + "_initial_chr",
        )

    for loop in range(30):  # update conformation F
        ## step1: derive Rotation Matrix R
        Mask = MASK(
            gem,
            GeneIDs=GeneUIDs,
            CellIDs=CellUIDs,
            Ngene=nderive,
        )
        RM, Z, CellUIDs, adata = DeriveRotation(Z, F, Mask, CellUIDs, adata)
        ## step2: update conformation F with R and Z
        try:
            newF, Z, GeneUIDs, F = UpdateF(RM, Z, GeneUIDs, F)
        except np.linalg.LinAlgError as e:
            print(
                f"{type(e).__name__} at line {e.__traceback__.tb_lineno} of {__file__}: {e}"
            )
            return "numpy.linalg.LinAlgError"
        rmsd = kabsch_numpy(
            normalizeF(F, method="mean"), normalizeF(newF, method="mean")
        )[2]
        print(
            "RMSD between New Configuration and Old Configuration for loop "
            + str(loop + 1)
            + " is: "
            + str(rmsd)
        )
        if outpath is not False:
            write_pdb(
                newF,
                gene_chr,
                GeneUIDs,
                write_path=outpath,
                sp=species,
                seed=seed,
                prefix=samplename + "_updated" + str(loop + 1) + "times_chr",
            )
        # renew conformation F
        F = newF

        ### test: save processing adata ###
        # adata.obsm["Rotation"] = RM
        # adata.uns["F"] = pd.DataFrame(F, index=GeneUIDs)
        # adata.uns["F"].columns = adata.uns["F"].columns.astype(str)
        # adata.uns["Z"] = Z
        # adata.uns["reconstruction_celllist"] = CellUIDs
        # adata.write_h5ad(filename=outpath + "loop_" + str(loop) + "_adata.h5ad")

        ## step3: check if conformation converges
        if rmsd < thresh_rmsd:
            break
    print("Number of total transcription centers is: " + str(F.shape[0]))

    adata.obsm["Rotation"] = RM
    adata.uns["F"] = pd.DataFrame(F, index=GeneUIDs)
    adata.uns["F"].columns = adata.uns["F"].columns.astype(str)
    adata.uns["Z"] = Z
    adata.uns["reconstruction_celllist"] = CellUIDs
    return adata


def parse_args():
    parser = argparse.ArgumentParser(description="Cytocraft Main Function")

    parser.add_argument(
        "-i",
        "--gem_path",
        type=str,
        help="Input path to gene expression matrix file",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--out_path",
        type=str,
        help="Directory to save results",
        required=True,
    )
    parser.add_argument(
        "--species",
        type=str,
        help="Species of the input data, e.g. Human, Mouse",
        choices=["Human", "Mouse", "Mice", "Axolotls", "Axolotl", "Monkey"],
        required=True,
    )
    parser.add_argument(
        "-p",
        "--percent",
        type=float,
        help="Percent of anchor gene, default: 0.001",
        default=None,
    )
    parser.add_argument(
        "-n",
        "--number",
        type=int,
        help="Number of anchor gene, recommend: 10",
        default=10,
    )
    parser.add_argument(
        "-t",
        "--gene_filter_thresh",
        type=float,
        help="The maximum allowable proportion of np.nan values in a column (representing a gene) of the observed transcription centers (Z), default: 0.90",
        default=0.90,
    )
    parser.add_argument(
        "-r",
        "--rmsd_thresh",
        type=float,
        help="RMSD threshold. If the computed RMSD value is less than or equal to this threshold, it means the process has reached an acceptable level of similarity or convergence, and the loop is exited. default: 0.01",
        default=0.01,
    )
    parser.add_argument(
        "--sep",
        type=str,
        help="Separator of the input gene expression matrix file",
        default="\t",
    )
    parser.add_argument(
        "-c",
        "--celltype",
        type=str,
        help="Path of obs file containing cell types",
    )
    parser.add_argument(
        "--ctkey",
        type=str,
        help="Key of celltype column in the annotation file",
    )
    parser.add_argument(
        "--cikey",
        type=str,
        help="Key of cell id column in the annotation file",
    )
    parser.add_argument(
        "--csep", type=str, help="Annotation file separator", default="\t"
    )
    parser.add_argument(
        "--seed",
        type=int,
        help="Random seed",
    )

    args = parser.parse_args()

    return args


def main():
    #################### SETTINGS ####################
    args = parse_args()
    if args.seed is not None:
        seed = args.seed
    else:
        seed = random.randint(0, 1000)

    ################ Check Gem Header ################
    gem_header = read_gem_header(args.gem_path, sep=args.sep)
    if not check_gem_header(gem_header):
        print(
            "Invalid gem header. Please make sure the gem file includes 'geneID', 'cell', 'x', 'y', 'MIDCount' columns."
        )
        return

    #################### RUNNING #####################
    if args.celltype is not None:  # run multiple cell types
        print("Running Multi-Celltype Mode")
        print("Cell types have been set (path is %s)" % args.celltype + "\n")
        # read gem and obs
        obs = pd.read_csv(args.celltype, sep=args.csep, dtype=str)
        # split gem
        split_paths = split_gem(
            args.gem_path,
            celltype=obs,
            ctkey=args.ctkey,
            cikey=args.cikey,
            gsep=args.sep,
        )
        # run cytocraft by order
        for ct, ct_path in split_paths.items():
            run_craft(
                gem_path=ct_path,
                species=args.species,
                seed=seed,
                out_path=args.out_path,
                sep="\t",
                threshold_for_gene_filter=args.gene_filter_thresh,
                threshold_for_rmsd=args.rmsd_thresh,
                Ngene_for_rotation_derivation=args.number,
                percent_of_gene_for_rotation_derivation=args.percent,
            )
            os.remove(ct_path)

    else:  # run single cell type
        print("Running Single-Celltype Mode")
        try:
            run_craft(
                gem_path=args.gem_path,
                species=args.species,
                seed=seed,
                out_path=args.out_path,
                sep=args.sep,
                threshold_for_gene_filter=args.gene_filter_thresh,
                threshold_for_rmsd=args.rmsd_thresh,
                Ngene_for_rotation_derivation=args.number,
                percent_of_gene_for_rotation_derivation=args.percent,
            )
        except Exception as e:
            print(
                f"{type(e).__name__} at line {e.__traceback__.tb_lineno} of {__file__}: {e}"
            )
            print("conformation reconstruction failed.")
        else:
            print("conformation reconstruction finished.")


if __name__ == "__main__":
    main()



import copy
import numpy as np
import pandas as pd
from anndata import AnnData
from typing import Optional, Union
from shapely.geometry import Point, MultiPoint
from pathlib import Path
from scipy.sparse import spmatrix, issparse, csr_matrix


class Cell(object):
    def __init__(
        self,
        cell_name: Optional[np.ndarray] = None,
        cell_border: Optional[np.ndarray] = None,
        batch: Optional[Union[np.ndarray, list, int, str]] = None,
    ):
        self._cell_name = cell_name
        self._cell_border = cell_border
        self._batch = self._set_batch(batch) if batch is not None else None
        self.total_counts = None
        self.pct_counts_mt = None
        self.n_genes_by_counts = None

    @property
    def cell_name(self):
        """
        get the name of cell.
        :return: cell name
        """
        return self._cell_name

    @cell_name.setter
    def cell_name(self, name: np.ndarray):
        """
        set the name of cell.
        :param name: a numpy array of names.
        :return:
        """
        if not isinstance(name, np.ndarray):
            raise TypeError("cell name must be a np.ndarray object.")
        self._cell_name = name

    @property
    def cell_border(self):
        return self._cell_border

    @cell_border.setter
    def cell_boder(self, cell_border: np.ndarray):
        if not isinstance(cell_border, np.ndarray):
            raise TypeError("cell border must be a np.ndarray object.")
        self._cell_border = cell_border

    @property
    def batch(self):
        return self._batch

    @batch.setter
    def batch(self, batch: Union[np.ndarray, list, int]):
        self._batch = self._set_batch(batch)

    def _set_batch(self, batch: Union[np.ndarray, list, int]):
        if batch is None:
            return None

        if (
            not isinstance(batch, np.ndarray)
            and not isinstance(batch, list)
            and not isinstance(batch, int)
            and not isinstance(batch, str)
        ):
            raise TypeError("batch must be np.ndarray or list or int or str")

        if isinstance(batch, int):
            batch = str(batch)
        if isinstance(batch, str):
            return np.repeat(batch, len(self.cell_name))
        else:
            return (np.array(batch) if isinstance(batch, list) else batch).astype("U")

    def sub_set(self, index):
        """
        get the subset of Cell by the index info, the Cell object will be inplaced by the subset.
        :param index: a numpy array of index info.
        :return: the subset of Cell object.
        """
        if self.cell_name is not None:
            self.cell_name = self.cell_name[index]
        if self.cell_boder is not None:
            self.cell_boder = self.cell_boder[index]
        if self.total_counts is not None:
            self.total_counts = self.total_counts[index]
        if self.pct_counts_mt is not None:
            self.pct_counts_mt = self.pct_counts_mt[index]
        if self.n_genes_by_counts is not None:
            self.n_genes_by_counts = self.n_genes_by_counts[index]
        if self.batch is not None:
            self.batch = self.batch[index]
        return self

    def get_property(self, name):
        """
        get the property value by the name.
        :param name: the name of property.
        :return: the property.
        """
        if name == "total_counts":
            return self.total_counts
        if name == "pct_counts_mt":
            return self.pct_counts_mt
        if name == "n_genes_by_counts":
            return self.n_genes_by_counts

    def to_df(self):
        """
        transform Cell object to pd.DataFrame.
        :return: a dataframe of Cell.
        """
        attributes = {
            "total_counts": self.total_counts,
            "pct_counts_mt": self.pct_counts_mt,
            "n_genes_by_counts": self.n_genes_by_counts,
        }
        if self._batch is not None:
            attributes["batch"] = self._batch
        df = pd.DataFrame(attributes, index=self.cell_name)
        return df


class AnnBasedCell(Cell):
    def __init__(
        self,
        based_ann_data: AnnData,
        cell_name: Optional[np.ndarray] = None,
        cell_border: Optional[np.ndarray] = None,
        batch: Optional[Union[np.ndarray, list, int, str]] = None,
    ):
        self.__based_ann_data = based_ann_data
        super(AnnBasedCell, self).__init__(cell_name, cell_border, batch)

    def __str__(self):
        return str(self.__based_ann_data.obs)

    def __repr__(self):
        return self.__str__()

    @property
    def cell_name(self) -> np.ndarray:
        """
        get the name of cell.
        :return: cell name
        """
        return self.__based_ann_data.obs_names.values.astype(str)

    @cell_name.setter
    def cell_name(self, name: np.ndarray):
        """
        set the name of cell.
        :param name: a numpy array of names.
        :return:
        """
        if not isinstance(name, np.ndarray):
            raise TypeError("cell name must be a np.ndarray object.")
        self.__based_ann_data._inplace_subset_obs(name)

    @property
    def total_counts(self):
        return self.__based_ann_data.obs["total_counts"]

    @total_counts.setter
    def total_counts(self, new_total_counts):
        if new_total_counts is not None:
            self.__based_ann_data.obs["total_counts"] = new_total_counts

    @property
    def pct_counts_mt(self):
        return self.__based_ann_data.obs["pct_counts_mt"]

    @pct_counts_mt.setter
    def pct_counts_mt(self, new_pct_counts_mt):
        if new_pct_counts_mt is not None:
            self.__based_ann_data.obs["pct_counts_mt"] = new_pct_counts_mt

    @property
    def n_genes_by_counts(self):
        return self.__based_ann_data.obs["n_genes_by_counts"]

    @n_genes_by_counts.setter
    def n_genes_by_counts(self, new_n_genes_by_counts):
        if new_n_genes_by_counts is not None:
            self.__based_ann_data.obs["n_genes_by_counts"] = new_n_genes_by_counts


class Gene(object):
    def __init__(self, gene_name: Optional[np.ndarray]):
        self._gene_name = gene_name if gene_name is None else gene_name.astype("U")
        self.n_cells = None
        self.n_counts = None

    @property
    def gene_name(self):
        """
        get the genes name.
        :return: genes name.
        """
        return self._gene_name

    @gene_name.setter
    def gene_name(self, name: np.ndarray):
        """
        set the name of gene.
        :param name: a numpy array of names.
        :return:
        """
        if not isinstance(name, np.ndarray):
            raise TypeError("gene name must be a np.ndarray object.")
        self._gene_name = name.astype("U")

    def sub_set(self, index):
        """
        get the subset of Gene by the index info, the Gene object will be inplaced by the subset.
        :param index: a numpy array of index info.
        :return: the subset of Gene object.
        """
        if self.gene_name is not None:
            self.gene_name = self.gene_name[index]
        if self.n_cells is not None:
            self.n_cells = self.n_cells[index]
        if self.n_counts is not None:
            self.n_counts = self.n_counts[index]
        return self

    def to_df(self):
        """
        transform Gene object to pd.DataFrame.
        :return: a dataframe of Gene.
        """
        attributes = {
            "n_counts": self.n_counts,
            "n_cells": self.n_cells,
        }
        df = pd.DataFrame(attributes, index=self.gene_name)
        return df


class AnnBasedGene(Gene):
    def __init__(self, based_ann_data: AnnData, gene_name: Optional[np.ndarray]):
        self.__based_ann_data = based_ann_data
        super(AnnBasedGene, self).__init__(gene_name)

    def __str__(self):
        return str(self.__based_ann_data.var)

    def __repr__(self):
        return self.__str__()

    @property
    def gene_name(self) -> np.ndarray:
        """
        get the genes name.
        :return: genes name.
        """
        return self.__based_ann_data.var_names.values.astype(str)

    @gene_name.setter
    def gene_name(self, name: np.ndarray):
        """
        set the name of gene.
        :param name: a numpy array of names.
        :return:
        """
        if not isinstance(name, np.ndarray):
            raise TypeError("gene name must be a np.ndarray object.")
        self.__based_ann_data._inplace_subset_var(name)


class Data(object):
    def __init__(
        self,
        file_path: Optional[str] = None,
        file_format: Optional[str] = None,
        output: Optional[str] = None,
        partitions: int = 1,
    ):
        self._file = Path(file_path) if file_path is not None else None
        self._partitions = int(partitions)
        self._file_format = file_format
        self.format_range = ["gem", "gef", "mtx", "h5ad", "scanpy_h5ad"]
        self._output = output

    def check(self):
        """
        checking whether the params is in the range.
        :return:
        """
        self.file_check(file=self.file)
        self.format_check(f_format=self.file_format)
        if self.file is not None and self.file_format is None:
            print("the file format must be not None , if the file path is set.")
            raise Exception

    @property
    def output(self):
        """
        the path of output.
        :return:
        """
        return self._output

    @output.setter
    def output(self, path):
        self.output_check(path)
        self._output = path

    @staticmethod
    def file_check(file):
        """
        Check if the file exists.
        :param file: the Path of file.
        :return:
        """
        if file is not None and not file.exists():
            print(f"{str(file)} is not exist, please check!")
            raise FileExistsError

    @property
    def file(self):
        """
        get the file property
        :return:
        """
        return self._file

    @file.setter
    def file(self, path):
        """
        set the file property
        :param path: the file path
        :return:
        """
        if path is None:
            file = path
        elif isinstance(path, str):
            file = Path(path)
        elif isinstance(path, Path):
            file = path
        else:
            raise TypeError
        self.file_check(file=file)
        self._file = file

    @property
    def file_format(self):
        """
        get the file_format property
        :return:
        """
        return self._file_format

    @file_format.setter
    def file_format(self, f_format):
        """
        set the file_format property
        :param f_format: the file format
        :return:
        """
        self.format_check(f_format)
        self._file_format = f_format

    @property
    def partitions(self):
        """
        get the partitions property
        :return:
        """
        return self._partitions

    @partitions.setter
    def partitions(self, partition):
        """
        set the partitions property
        :param partition: the partitions number, which define the cores of multi processes.
        :return:
        """
        self._partitions = partition


class StereoExpData(Data):
    def __init__(
        self,
        file_path: Optional[str] = None,
        file_format: Optional[str] = None,
        bin_type: Optional[str] = None,
        bin_size: int = 100,
        exp_matrix: Optional[Union[np.ndarray, spmatrix]] = None,
        genes: Optional[Union[np.ndarray, Gene]] = None,
        cells: Optional[Union[np.ndarray, Cell]] = None,
        position: Optional[np.ndarray] = None,
        output: Optional[str] = None,
        partitions: int = 1,
        offset_x: Optional[str] = None,
        offset_y: Optional[str] = None,
        attr: Optional[dict] = None,
        merged: bool = False,
    ):
        """
        a Data designed for express matrix of spatial omics. It can directly set the corresponding properties
        information to initialize the data. If the file path is not None, we will read the file information to
        initialize the properties.
        :param file_path: the path of express matrix file.
        :param file_format: the file format of the file_path.
        :param bin_type: the type of bin, if file format is stereo-seq file. `bins` or `cell_bins`.
        :param bin_size: size of bin to merge if bin type is 'bins'.
        :param exp_matrix: the express matrix.
        :param genes: the gene object which contain some info of gene.
        :param cells: the cell object which contain some info of cell.
        :param position: the spatial location.
        :param output: the path of output.
        :param partitions: the number of multi-process cores, used when processing files in parallel.
        :param offset_x: the x of the offset.
        :param offset_y: the y of the offset.
        :param attr: attributions from gef file.
        """
        super(StereoExpData, self).__init__(
            file_path=file_path,
            file_format=file_format,
            partitions=partitions,
            output=output,
        )
        self._exp_matrix = exp_matrix
        self._genes = genes if isinstance(genes, Gene) else Gene(gene_name=genes)
        self._cells = cells if isinstance(cells, Cell) else Cell(cell_name=cells)
        self._position = position
        self._position_offset = None
        self._bin_type = bin_type
        self.bin_size = bin_size
        self._tl = None
        self._plt = None
        self.raw = None
        self._offset_x = offset_x
        self._offset_y = offset_y
        self._attr = attr
        self._merged = merged
        self._sn = self.get_sn_from_path(file_path)

    def get_sn_from_path(self, file_path):
        if file_path is None:
            return None

        from os import path

        return path.basename(file_path).split(".")[0].strip()

    @property
    def tl(self):
        if self._tl is None:
            self._tl = StPipeline(self)
        return self._tl

    def sub_by_index(self, cell_index=None, gene_index=None):
        """
        get sub data by cell index or gene index list.
        :param cell_index: a list of cell index.
        :param gene_index: a list of gene index.
        :return:
        """
        if cell_index is not None:
            self.exp_matrix = self.exp_matrix[cell_index, :]
            self.position = (
                self.position[cell_index, :] if self.position is not None else None
            )
            self.cells = self.cells.sub_set(cell_index)
        if gene_index is not None:
            self.exp_matrix = self.exp_matrix[:, gene_index]
            self.genes = self.genes.sub_set(gene_index)
        return self

    def sub_by_name(
        self,
        cell_name: Optional[Union[np.ndarray, list]] = None,
        gene_name: Optional[Union[np.ndarray, list]] = None,
    ):
        """
        get sub data by cell name list or gene name list.
        :param cell_name: a list of cell name.
        :param gene_name: a list of gene name.
        :return:
        """
        data = copy.deepcopy(self)
        cell_index = (
            [np.argwhere(data.cells.cell_name == i)[0][0] for i in cell_name]
            if cell_name is not None
            else None
        )
        gene_index = (
            [np.argwhere(data.genes.gene_name == i)[0][0] for i in gene_name]
            if gene_name is not None
            else None
        )
        return data.sub_by_index(cell_index, gene_index)

    def check(self):
        """
        checking whether the params is in the range.
        :return:
        """
        super(StereoExpData, self).check()
        self.bin_type_check(self._bin_type)

    @staticmethod
    def bin_type_check(bin_type):
        """
        check whether the bin type is in range.
        :param bin_type: bin type value, 'bins' or 'cell_bins'.
        :return:
        """
        if (bin_type is not None) and (bin_type not in ["bins", "cell_bins"]):
            print(f"the bin type `{bin_type}` is not in the range, please check!")
            raise Exception

    @property
    def shape(self):
        return self.exp_matrix.shape

    @property
    def gene_names(self):
        """
        get the gene names.
        :return:
        """
        return self.genes.gene_name

    @property
    def cell_names(self):
        """
        get the cell names.
        :return:
        """
        return self.cells.cell_name

    @property
    def cell_borders(self):
        return self.cells.cell_boder

    @property
    def genes(self):
        """
        get the value of self._genes.
        :return:
        """
        return self._genes

    @genes.setter
    def genes(self, gene):
        """
        set the value of self._genes.
        :param gene: a object of Gene
        :return:
        """
        self._genes = gene

    @property
    def cells(self):
        """
        get the value of self._cells
        :return:
        """
        return self._cells

    @cells.setter
    def cells(self, cell):
        """
        set the value of self._cells.
        :param cell: a object of Cell
        :return:
        """
        self._cells = cell

    @property
    def exp_matrix(self):
        """
        get the value of self._exp_matrix.
        :return:
        """
        return self._exp_matrix

    @exp_matrix.setter
    def exp_matrix(self, pos_array):
        """
        set the value of self._exp_matrix.
        :param pos_array: np.ndarray or sparse.spmatrix.
        :return:
        """
        self._exp_matrix = pos_array

    @property
    def bin_type(self):
        """
        get the value of self._bin_type.
        :return:
        """
        return self._bin_type

    @bin_type.setter
    def bin_type(self, b_type):
        """
        set the value of self._bin_type.
        :param b_type: the value of bin type, 'bins' or 'cell_bins'.
        :return:
        """
        self.bin_type_check(b_type)
        self._bin_type = b_type

    @property
    def position(self):
        """
        get the value of self._position.
        :return:
        """
        return self._position

    @position.setter
    def position(self, pos):
        """
        set the value of self._position.
        :param pos: the value of position, a np.ndarray .
        :return:
        """
        self._position = pos

    @property
    def position_offset(self):
        return self._position_offset

    @position_offset.setter
    def position_offset(self, position_offset):
        self._position_offset = position_offset

    @property
    def offset_x(self):
        """
        get the x of self._offset_x.
        :return:
        """
        return self._offset_x

    @offset_x.setter
    def offset_x(self, min_x):
        """
        :param min_x: offset of x.
        :return:
        """
        self._offset_x = min_x

    @property
    def offset_y(self):
        """
        get the offset_y of self._offset_y.
        :return:
        """
        return self._offset_y

    @offset_y.setter
    def offset_y(self, min_y):
        """
        :param min_y: offset of y.
        :return:
        """
        self._offset_y = min_y

    @property
    def attr(self):
        """
        get the attr of self._attr.
        :return:
        """
        return self._attr

    @attr.setter
    def attr(self, attr):
        """
        :param attr: dict of attr.
        :return:
        """
        self._attr = attr

    @property
    def merged(self):
        return self._merged

    @merged.setter
    def merged(self, merged):
        self._merged = merged

    @property
    def sn(self):
        return self._sn

    @sn.setter
    def sn(self, sn):
        self._sn = sn

    def to_df(self):
        """
        transform StereoExpData to pd.DataFrame.
        :return:
        """
        df = pd.DataFrame(
            self.exp_matrix.toarray() if issparse(self.exp_matrix) else self.exp_matrix,
            columns=self.gene_names,
            index=self.cell_names,
        )
        return df

    def sparse2array(self):
        """
        transform expression matrix to array if it is parse matrix.
        :return:
        """
        if issparse(self.exp_matrix):
            self.exp_matrix = self.exp_matrix.toarray()
        return self.exp_matrix

    def array2sparse(self):
        """
        transform expression matrix to sparse matrix if it is ndarray
        :return:
        """
        if not issparse(self.exp_matrix):
            self.exp_matrix = csr_matrix(self.exp_matrix)
        return self.exp_matrix

    def __str__(self):
        format_str = f"StereoExpData object with n_cells X n_genes = {self.shape[0]} X {self.shape[1]}"
        format_str += f"\nbin_type: {self.bin_type}"
        if self.bin_type == "bins":
            format_str += f"\n{'bin_size: %d' % self.bin_size}"
        format_str += f"\noffset_x = {self.offset_x}"
        format_str += f"\noffset_y = {self.offset_y}"
        format_cells = []
        for attr_name in [
            ("_cell_name", "cell_name"),
            "total_counts",
            "n_genes_by_counts",
            "pct_counts_mt",
        ]:
            if type(attr_name) is tuple:
                real_name, show_name = attr_name[0], attr_name[1]
            else:
                real_name = show_name = attr_name
            # `is not None` is ugly but object in __dict__ may be a pandas.DataFrame
            if self.cells.__dict__.get(real_name, None) is not None:
                format_cells.append(show_name)
        if format_cells:
            format_str += f"\ncells: {format_cells}"
        format_genes = []
        for attr_name in [("_gene_name", "gene_name"), "n_counts", "n_cells"]:
            if type(attr_name) is tuple:
                real_name, show_name = attr_name[0], attr_name[1]
            else:
                real_name = show_name = attr_name
            if self.genes.__dict__.get(real_name, None) is not None:
                format_genes.append(show_name)
        if format_genes:
            format_str += f"\ngenes: {format_genes}"
        # TODO: no decide yet
        # format_str += "\nposition: T"
        format_key_record = {
            key: value for key, value in self.tl.key_record.items() if value
        }
        if format_key_record:
            format_str += f"\nkey_record: {format_key_record}"
        return format_str

    def __repr__(self):
        return self.__str__()

    def issparse(self):
        return issparse(self.exp_matrix)


class StPipeline(object):
    def __init__(self, data):
        """
        A analysis tool sets for StereoExpData. include preprocess, filter, cluster, plot and so on.
        :param data: StereoExpData object.
        """
        self.data = data
        self.result = dict()
        self._raw = None
        self.key_record = {
            "hvg": [],
            "pca": [],
            "neighbors": [],
            "umap": [],
            "cluster": [],
            "marker_genes": [],
        }

    @property
    def raw(self):
        """
        get the StereoExpData whose exp_matrix is raw count.
        :return:
        """
        return self._raw

    @raw.setter
    def raw(self, value):
        """
        set the raw data.
        :param value: StereoExpData.
        :return:
        """
        self._raw = copy.deepcopy(value)

    def reset_raw_data(self):
        """
        reset the self.data to the raw data.
        :return:
        """
        self.data = self.raw

    def raw_checkpoint(self):
        """
        Save the current results to self.raw.
        :param key:
        :param res_key:
        :return:
        """
        self.raw = self.data

    def reset_key_record(self, key, res_key):
        """
        reset key and coordinated res_key in key_record.
        :param key:
        :param res_key:
        :return:
        """
        if key in self.key_record.keys():
            if res_key in self.key_record[key]:
                self.key_record[key].remove(res_key)
            self.key_record[key].append(res_key)
        else:
            self.key_record[key] = [res_key]


def read_gem(file_path, sep="\t", bin_type="cell_bins", bin_size=100, is_sparse=True):
    """
    read the stereo-seq file, and generate the object of StereoExpData.
    :param file_path: input file
    :param sep: separator string
    :param bin_type: the type of bin, if file format is stereo-seq file. `bins` or `cell_bins`.
    :param bin_size: the size of bin to merge. The parameter only takes effect
                    when the value of data.bin_type is 'bins'.
    :param is_sparse: the matrix is sparse matrix if is_sparse is True else np.ndarray
    :return: an object of StereoExpData.
    """
    data = StereoExpData(file_path=file_path, bin_type=bin_type, bin_size=bin_size)
    df = pd.read_csv(
        str(data.file),
        sep=sep,
        comment="#",
        header=0,
        converters={
            "x": int,
            "y": int,
            "MIDCount": float,
            "MIDCounts": float,
            "expr": float,
        },
    )
    # counts
    if "MIDCounts" in df.columns:
        df.rename(columns={"MIDCounts": "UMICount"}, inplace=True)
    elif "MIDCount" in df.columns:
        df.rename(columns={"MIDCount": "UMICount"}, inplace=True)
    elif "expr" in df.columns:
        df.rename(columns={"expr": "UMICount"}, inplace=True)
    # cells
    if "CellID" in df.columns:
        df.rename(columns={"CellID": "cell_id"}, inplace=True)
    elif "cell" in df.columns:
        df.rename(columns={"cell": "cell_id"}, inplace=True)
    elif "label" in df.columns:
        df.rename(columns={"label": "cell_id"}, inplace=True)
    # genes
    if "gene" in df.columns:
        df.rename(columns={"gene": "geneID"}, inplace=True)
    df.dropna(inplace=True)
    gdf = None
    gdf = parse_cell_bin_coor(df)
    cells = df["cell_id"].unique()
    genes = df["geneID"].unique()
    cells_dict = dict(zip(cells, range(0, len(cells))))
    genes_dict = dict(zip(genes, range(0, len(genes))))
    rows = df["cell_id"].map(cells_dict)
    cols = df["geneID"].map(genes_dict)
    exp_matrix = csr_matrix(
        (df["UMICount"], (rows, cols)),
        shape=(cells.shape[0], genes.shape[0]),
        dtype=np.float32,
    )
    data.cells = Cell(cell_name=cells)
    data.genes = Gene(gene_name=genes)
    data.exp_matrix = exp_matrix if is_sparse else exp_matrix.toarray()
    data.position = gdf.loc[cells][["x_center", "y_center"]].values
    data.cells.cell_point = gdf.loc[cells]["cell_point"].values
    data.offset_x = df["x"].min()
    data.offset_y = df["y"].min()
    data.attr = {
        "minX": df["x"].min(),
        "minY": df["y"].min(),
        "maxX": df["x"].max(),
        "maxY": df["y"].max(),
        "minExp": data.exp_matrix.toarray().min()
        if is_sparse
        else data.exp_matrix.min(),
        "maxExp": data.exp_matrix.toarray().max()
        if is_sparse
        else data.exp_matrix.min(),
        "resolution": 0,
    }
    return data


def stereo_to_anndata(
    data: StereoExpData, flavor="scanpy", sample_id="sample", reindex=False, output=None
):
    """
    transform the StereoExpData object into Anndata object.
    :param data: StereoExpData object
    :param flavor: 'scanpy' or 'seurat'.
    if you want to convert the output_h5ad into h5seurat for seurat, please set 'seurat'.
    :param sample_id: sample name, which will be set as 'orig.ident' in obs.
    :param reindex: if True, the cell index will be reindex as "{sample_id}:{position_x}_{position_y}" format.
    :param output: path of output_file(.h5ad).
    :return: Anndata object
    """
    from scipy.sparse import issparse

    if data.tl.raw is None:
        print("convert to AnnData should have raw data")
        raise Exception

    exp = (
        data.tl.raw.exp_matrix
        if issparse(data.tl.raw.exp_matrix)
        else csr_matrix(data.tl.raw.exp_matrix)
    )
    cells = data.tl.raw.cells.to_df()
    cells.dropna(axis=1, how="all", inplace=True)
    genes = data.tl.raw.genes.to_df()
    genes.dropna(axis=1, how="all", inplace=True)

    adata = AnnData(
        X=exp,
        dtype=np.float64,
        obs=cells,
        var=genes,
        # uns={'neighbors': {'connectivities_key': 'None','distance_key': 'None'}},
    )
    ##sample id
    print(f"Adding {sample_id} in adata.obs['orig.ident'].")
    adata.obs["orig.ident"] = pd.Categorical(
        [sample_id] * adata.obs.shape[0], categories=[sample_id]
    )
    if data.position is not None:
        print(f"Adding data.position as adata.obsm['spatial'] .")
        adata.obsm["spatial"] = data.position
        # adata.obsm['X_spatial'] = data.position
        print(f"Adding data.position as adata.obs['x'] and adata.obs['y'] .")
        adata.obs["x"] = pd.DataFrame(
            data.position[:, 0], index=data.cell_names.astype("str")
        )
        adata.obs["y"] = pd.DataFrame(
            data.position[:, 1], index=data.cell_names.astype("str")
        )

    if data.sn is not None:
        if isinstance(data.sn, str):
            sn_list = [["-1", data.sn]]
        else:
            sn_list = []
            for bno, sn in data.sn.items():
                sn_list.append([bno, sn])
        adata.uns["sn"] = pd.DataFrame(sn_list, columns=["batch", "sn"])

    for key in data.tl.key_record.keys():
        if len(data.tl.key_record[key]) > 0:
            if key == "hvg":
                res_key = data.tl.key_record[key][-1]
                print(f"Adding data.tl.result['{res_key}'] in adata.var .")
                adata.uns[key] = {"params": {}, "source": "stereopy", "method": key}
                for i in data.tl.result[res_key]:
                    if i == "mean_bin":
                        continue
                    adata.var[i] = data.tl.result[res_key][i]
            elif key == "sct":
                res_key = data.tl.key_record[key][-1]
                # adata.uns[res_key] = {}
                print(f"Adding data.tl.result['{res_key}'] in adata.uns['sct_'] .")
                adata.uns["sct_counts"] = csr_matrix(
                    data.tl.result[res_key][0]["counts"].T
                )
                adata.uns["sct_data"] = csr_matrix(data.tl.result[res_key][0]["data"].T)
                adata.uns["sct_scale"] = csr_matrix(
                    data.tl.result[res_key][0]["scale.data"].T.to_numpy()
                )
                adata.uns["sct_top_features"] = list(
                    data.tl.result[res_key][0]["scale.data"].index
                )
                adata.uns["sct_cellname"] = list(
                    data.tl.result[res_key][1]["umi_cells"].astype("str")
                )
                adata.uns["sct_genename"] = list(
                    data.tl.result[res_key][1]["umi_genes"]
                )
            elif key in ["pca", "umap", "tsne"]:
                # pca :we do not keep variance and PCs(for varm which will be into feature.finding in pca of seurat.)
                res_key = data.tl.key_record[key][-1]
                sc_key = f"X_{key}"
                print(f"Adding data.tl.result['{res_key}'] in adata.obsm['{sc_key}'] .")
                adata.obsm[sc_key] = data.tl.result[res_key].values
            elif key == "neighbors":
                # neighbor :seurat use uns for conversion to @graph slot, but scanpy canceled neighbors of uns at present.
                # so this part could not be converted into seurat straightly.
                for res_key in data.tl.key_record[key]:
                    sc_con = (
                        "connectivities"
                        if res_key == "neighbors"
                        else f"{res_key}_connectivities"
                    )
                    sc_dis = (
                        "distances"
                        if res_key == "neighbors"
                        else f"{res_key}_distances"
                    )
                    print(
                        f"Adding data.tl.result['{res_key}']['connectivities'] in adata.obsp['{sc_con}'] ."
                    )
                    print(
                        f"Adding data.tl.result['{res_key}']['nn_dist'] in adata.obsp['{sc_dis}'] ."
                    )
                    adata.obsp[sc_con] = data.tl.result[res_key]["connectivities"]
                    adata.obsp[sc_dis] = data.tl.result[res_key]["nn_dist"]
                    print(f"Adding info in adata.uns['{res_key}'].")
                    adata.uns[res_key] = {}
                    adata.uns[res_key]["connectivities_key"] = sc_con
                    adata.uns[res_key]["distance_key"] = sc_dis
                    # adata.uns[res_key]['connectivities'] = data.tl.result[res_key]['connectivities']
                    # adata.uns[res_key]['distances'] = data.tl.result[res_key]['nn_dist']
            elif key == "cluster":
                for res_key in data.tl.key_record[key]:
                    print(
                        f"Adding data.tl.result['{res_key}'] in adata.obs['{res_key}'] ."
                    )
                    adata.obs[res_key] = pd.DataFrame(
                        data.tl.result[res_key]["group"].values,
                        index=data.cells.cell_name.astype("str"),
                    )
            elif key == "gene_exp_cluster":
                for res_key in data.tl.key_record[key]:
                    adata.uns[res_key] = data.tl.result[res_key]
            else:
                continue

        if reindex:
            print(f"Reindex adata.X .")
            new_ix = (
                adata.obs["orig.ident"].astype(str)
                + ":"
                + adata.obs["x"].astype(str)
                + "_"
                + adata.obs["y"].astype(str)
            ).tolist()
            adata.obs.index = new_ix
            if "sct_cellname" in adata.uns.keys():
                print(f"Reindex as adata.uns['sct_cellname'] .")
                adata.uns["sct_cellname"] = new_ix

        print(f"Finished conversion to anndata.")

        if output is not None:
            adata.write_h5ad(output)
            print(f"Finished output to {output}")

        return adata


def parse_cell_bin_coor(df):
    gdf = df.groupby("cell_id").apply(lambda x: make_multipoint(x))
    return gdf


def make_multipoint(x):
    p = [Point(i) for i in zip(x["x"], x["y"])]
    mlp = MultiPoint(p).convex_hull
    x_center = mlp.centroid.x
    y_center = mlp.centroid.y
    return pd.Series({"cell_point": mlp, "x_center": x_center, "y_center": y_center})



from math import cos, sin
import numpy as np
from numpy.linalg import svd, solve, lstsq
from cytocraft import model
from cytocraft import util
from cytocraft.model import BasisShapeModel

"""
This module implements rigid factorization.
"""


def is_rotation_matrix(matrix):
    # Check if the matrix is square and has shape (3, 3)
    if not isinstance(matrix, np.ndarray) or matrix.shape != (3, 3):
        print("error1")
        return False

    # Check if the matrix is orthogonal
    if not np.allclose(matrix.T @ matrix, np.eye(3), atol=0.0001):
        print("error2")
        return False

    # Check if the matrix has determinant 1
    if not np.isclose(np.linalg.det(matrix), 1, atol=0.0001):
        print("error3")
        return False

    # Check if the matrix preserves the length of a random vector
    v = np.random.rand(3)
    if not np.isclose(np.linalg.norm(matrix @ v), np.linalg.norm(v), atol=0.0001):
        print("error4")
        return False

    # If all checks pass, return True
    return True


def factor_matrix(W, J=3):
    """Obtain the best rank J factorization MS to W

    Computes a factorization of W into M and B such that ||W-MB||_F
    is minimized.

    Input:
    W is a MxN matrix.

    Returns:
    M -- A MxR matrix
    B -- A RxN matrix
    """

    # Decompose using SVD.
    U, s, Vt = svd(W, full_matrices=False)

    # Compute the factorization.
    sqrt_sigma = np.diag(np.sqrt(s[:J]))
    M = np.dot(U[:, :J], sqrt_sigma)
    B = np.dot(sqrt_sigma, Vt[:J])

    return M, B


def factor(W):
    """
    This implements rigid factorization as described in

    Tomasi, C. & Kanade, T. "Shape and motion from image streams under
    orthography: a factorization method International Journal of Computer
    Vision, 1992
    """

    F = int(W.shape[0] / 2)
    N = W.shape[1]

    # Center W
    T = W.mean(axis=1)
    W = W - T[:, np.newaxis]

    # Factor W
    M_hat, B_hat = factor_matrix(W, J=3)

    # Where we will build the linear system.
    A = np.zeros((3 * F, 6))
    b = np.zeros((3 * F,))

    for f in range(F):
        # Extract the two rows.
        x_f, y_f = M_hat[f * 2 : f * 2 + 2]

        # Both rows must have unit length.
        A[f * 3] = util.vc(x_f, x_f)
        b[f * 3] = 1.0
        A[f * 3 + 1] = util.vc(y_f, y_f)
        b[f * 3 + 1] = 1.0

        # And be orthogonal.
        A[f * 3 + 2] = util.vc(x_f - y_f, x_f + y_f)

    # Recovec vech(Q) and Q
    vech_Q = lstsq(A, b, rcond=None)[0]
    Q = util.from_vech(vech_Q, 3, 3, sym=True)

    # Factor out G recovery matrix
    G, Gt = factor_matrix(Q)

    # Upgrade M and B.
    M = np.dot(M_hat, G)
    B = solve(G, B_hat)

    # Find actual rotations matrices.
    Rs = np.zeros((F, 3, 3))
    Rs[:, :2] = M.reshape(F, 2, 3)
    Rs[:, 2] = np.cross(Rs[:, 0], Rs[:, 1], axis=-1)
    # Rs[:, 2] = util.normed(np.cross(Rs[:, 0], Rs[:, 1], axis=-1))

    # check reflection
    # M = np.einsum("ijk, kl->ijl", Rs, np.linalg.inv(G))
    # is_reflection = (np.linalg.det(M) * np.linalg.det(G)) < 0.0
    # print(is_reflection)
    # I = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -1]])
    # if is_reflection.any():
    #     print("reflect")
    #     Rs = np.einsum("ijk, kl->ijl", np.einsum("jk, ikl->ijl", I, Rs), I)

    # And 3D translations.
    Ts = np.zeros((F, 3))
    Ts[:, :2] = T.reshape(F, 2)

    model = BasisShapeModel(Rs, Bs=B[np.newaxis, :, :], Ts=Ts)

    return model



"""
    Utility functions.
"""

import numpy as np
import random, pickle


def vec(X, order="F"):
    """Returns the vectorization of X.  Columns of X are stacked.  (The opposite of X.flatten())."""
    assert X.ndim == 2, "vec operator requires a matrix."

    return X.flatten(order=order)


def from_vec(x, m, n, order="F"):
    return x.reshape(m, n, order=order)


def vech(X, order="F"):
    """Returns vectorization of lower triangle of X in column major order by default."""
    assert X.ndim == 2, "vech operator requires a matrix."

    m, n = X.shape
    if order == "F":
        idx = np.where(1 - np.tri(n, m, -1, dtype=int))
        return X.T[idx]
    elif order == "C":
        i, j = np.where(np.tri(m, n, dtype=int))
    else:
        raise Exception("Only order C and F are allowed")

    return X[i, j]


def from_vech(x, m, n, order="F", sym=False):
    X = np.zeros((m, n), dtype=x.dtype)
    if order == "F":
        idx = np.where(1 - np.tri(n, m, -1, dtype=int))
        X.T[idx] = x
    elif order == "C":
        idx = np.where(np.tri(m, n, dtype=int))
        X[idx] = x
    else:
        raise Exception("Only C and F ordering allowed.")

    if sym:
        if m != n:
            raise Exception("Can only do this for square matrices.")
        X = X + X.T - np.diag(np.diag(X))
    return X


def vc(a, b):
    """vc operator from Brand 2005"""
    assert a.ndim == 1 and b.ndim == 1 and a.size == b.size
    a = a[:, np.newaxis]
    b = b[:, np.newaxis]
    return vech(np.dot(a, b.T) + np.dot(b, a.T) - np.diag((a * b).flat))


def hat_operator(omega):
    F = omega.shape[0]

    # Create the hat matrix.
    OH = np.zeros((F, 3, 3))

    o1, o2, o3 = omega.T

    OH[:, 0, 1] = -o3
    OH[:, 0, 2] = o2
    OH[:, 1, 0] = o3
    OH[:, 1, 2] = -o1
    OH[:, 2, 0] = -o2
    OH[:, 2, 1] = o1

    return OH


def axis_angle_to_Rs(omega, theta):
    F = theta.shape[0]

    # Calculate omega hat.
    omega_hat = hat_operator(omega)

    # Use Rodriguez' formula
    Rs = np.zeros((F, 3, 3))

    # Term 1.
    Rs += np.eye(3)[np.newaxis, ...]

    # Term 2.
    Rs += np.sin(theta)[:, np.newaxis, np.newaxis] * omega_hat

    # Term 3.
    tmp = omega[:, :, np.newaxis] * omega[:, np.newaxis, :] - np.eye(3)[np.newaxis, ...]
    Rs += (1 - np.cos(theta))[:, np.newaxis, np.newaxis] * tmp

    return Rs


def rotvecs_to_Rs(omega):
    """
    omega - Fx3 rotation vectors.

    Returns a Fx3x3 Rs tensor.
    """

    # Make sure, that we don't modify the original vector.
    omega = omega.copy()

    # Calculate the norm of the vectors.
    theta = norm(omega, axis=1)

    # When theta is zero, the rotation will be the identity
    # and 2pi is also the identity.
    theta[theta == 0] = np.pi * 2

    # This allows us to normalize without getting NaNs
    omega /= theta[:, np.newaxis]

    # Now use Rodrigues' formula to calculate
    Rs = axis_angle_to_Rs(omega, theta)

    return Rs


def norm(x, axis=-1):
    return np.sqrt((x * x).sum(axis=axis))


def rms(x, axis=None):
    return np.sqrt(np.mean(np.square(x), axis=axis))


def normed(x, axis=-1):
    if axis < -x.ndim or axis >= x.ndim:
        raise ValueError("axis(=%d) out of bounds" % axis)
    if axis < 0:
        axis += x.ndim
    shape = list(x.shape)
    shape[axis] = 1
    shape = tuple(shape)
    return x / norm(x, axis=axis).reshape(shape)



import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)
import pickle, copy, importlib, math, os
import seaborn as sns
import numpy as np
import pandas as pd
import umap
import scanpy as sc
from tqdm import tqdm
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from scipy.cluster.hierarchy import fcluster

# seed = np.random.RandomState(seed=3)


def get_cmap(cmap, ncolor):
    # Define the color map
    cmap = plt.get_cmap(cmap)

    # Get the RGB values of each color in the color map
    hex_values = []
    for i in range(ncolor):
        hex_values.append(mcolors.rgb2hex(cmap(int(256 / ncolor) * i)))

    return hex_values


def plot_RMSD_heatmap(
    D,
    labels,
    cmap="viridis",
    method="average",
    cthresh=4.8,
    font_scale=1,
    legend=True,
    cbbox=(0.07, 0.84, 0.04, 0.12),
    lgbox=(0.86, 0.96, 0.01, 0.01),
    ticks=[0, 0.5, 1, 1.5],
    figsize=(8, 8),
    dpi=300,
    legend_ncol=2,
    mask=None,
):
    import matplotlib.pyplot as plt

    clustergrid = sns.clustermap(
        D, xticklabels=labels, yticklabels=labels, method=method
    )
    plt.close()

    # Get clusters with threshold
    linkage = clustergrid.dendrogram_row.linkage
    threshold = cthresh
    clusters = fcluster(linkage, threshold, criterion="distance")

    # Create a color map for clusters
    n_clusters = len(np.unique(clusters))
    color_map = sns.color_palette("rainbow", n_clusters)

    # Assign colors to clusters
    cluster_colors = [color_map[c - 1] for c in clusters]

    # Draw plot
    sns.set(rc={"figure.dpi": dpi, "figure.figsize": figsize}, font_scale=font_scale)

    ax = sns.clustermap(
        D,
        xticklabels=labels,
        yticklabels=labels,
        cmap=cmap,
        method=method,
        row_colors=[cluster_colors],
        col_colors=[cluster_colors],
        cbar_kws={"ticks": ticks},
        mask=mask,
    )
    ax.ax_cbar.set_title("RMSD")
    ax.ax_cbar.set_position(cbbox)

    if legend:
        # Add legend for row color bar
        from matplotlib.patches import Patch

        lut = dict(zip(clusters, cluster_colors))
        handles = [Patch(facecolor=lut[name]) for name in sorted(lut)]
        legend = plt.legend(
            handles,
            sorted(lut),
            title="Groups",
            bbox_to_anchor=lgbox,
            bbox_transform=plt.gcf().transFigure,
            loc="upper right",
            ncol=legend_ncol,
        )
        legend.get_frame().set_facecolor("white")

    return ax


def plot_network(
    D,
    labels,
    cmap="viridis",
    csep="\n",
    solver="barnesHut",
    corder=0,
    N_neighbor=3,
    html="test.html",
    fontsize=15,
    width=1400,
    height=1000,
    edge_scale=150,
    edge_adjust=-60,
):
    from pyvis.network import Network
    from IPython.display import HTML
    import plotly.graph_objs as go
    import itertools
    import networkx as nx
    import plotly.io as pio
    import kaleido
    import json

    # create an empty graph
    G = nx.Graph()

    # add nodes
    N = D.shape[0]
    for i in range(N):
        G.add_node(i, label=labels[i])

    # add edges
    edges = []

    for i in range(N):
        arr = D[i, :]
        min_indices = np.argpartition(arr, int(N_neighbor + 1))[: int(N_neighbor + 1)]
        min_values = arr[min_indices]
        for j in min_indices:
            if not i == j:
                edges.append(
                    (
                        i,
                        int(j),
                        {"weight": float(1 / arr[j] * edge_scale + edge_adjust)},
                    )
                )

    G.add_edges_from(edges)

    # Draw pyvis
    net = Network(width=width, height=height, notebook=True, cdn_resources="remote")
    net.from_nx(G)

    # set node color based on label
    for node in net.nodes:
        node["group"] = node["label"].split(csep)[corder]
        node["color"] = cmap[node["group"]]
        # print(node['color'])

    # show
    net.show_buttons()

    json_obj = {
        "configure": {"enabled": True, "filter": ["nodes", "edges", "physics"]},
        "nodes": {
            "borderWidth": 3,
            "opacity": 1,
            "font": {"size": fontsize, "strokeWidth": 5},
            "size": 0,
        },
        "edges": {
            "color": {"opacity": 0.7},
            "selfReferenceSize": 0,
            "selfReference": {"size": 0, "angle": 0.7853981633974483},
            "smooth": {"forceDirection": "vertical"},
        },
        "physics": {
            "minVelocity": 0.75,
            "solver": solver,
        },
    }

    net.set_options("const options = " + json.dumps(json_obj))

    net.show(html)


# UMAP
def umap_leiden(
    distance_matrix,
    n_neighbors=15,
    min_dist=0.1,
    resolution=1,
    plot_umap=True,
    seed=99,
    title="",
):
    umap_coord = umap.UMAP(
        metric="precomputed",
        n_neighbors=n_neighbors,
        min_dist=min_dist,
        random_state=seed,
    ).fit_transform(distance_matrix)
    adata = sc.AnnData(pd.DataFrame(np.zeros((len(umap_coord), len(umap_coord)))))
    adata.obsm["umap"] = umap_coord
    sc.pp.neighbors(adata, use_rep="umap")
    sc.tl.leiden(adata, resolution=resolution)
    if plot_umap:
        sc.pl.umap(adata, color=["leiden"], title=title)
    return adata.obs["leiden"]
